import pyfastx
from typing import Dict, Tuple


def parse_fasta_file(fasta_file: str) -> pyfastx.Fasta:
    """
    Parse a FASTA file and return a pyfastx.Fasta object.

    :param fasta_file: The path to the FASTA file.

    :return: A pyfastx.Fasta object representing the parsed FASTA file.
    """
    fa = pyfastx.Fasta(fasta_file, build_index=True)
    return fa


def make_contig_index(contigs: list) -> Tuple[Dict[str, int], Dict[int, str]]:
    """
    Create an index mapping for contigs.

    :param contigs: A list of contig names.

    :return: A tuple containing the contig index mapping dictionaries (contig_to_index, index_to_contig).
    """
    contig_to_index = {contig: index for index, contig in enumerate(contigs)}
    index_to_contig = {index: contig for contig, index in contig_to_index.items()}
    return contig_to_index, index_to_contig


def apply_contig_index(contig_to_index: Dict[str, int], contig_to_info: Dict[str, str]) -> Dict[int, str]:
    """
    Apply the contig index mapping to the contig info dictionary.

    :param contig_to_index: A dictionary mapping contig names to their corresponding index.
    :param contig_to_info: A dictionary mapping contig names to their associated information.

    :return: A dictionary mapping contig indices to their associated information.
    """
    return {contig_to_index[contig]: info for contig, info in contig_to_info.items()}



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from binette import __version__

project = 'Binette'
copyright = '2024, Jean Mainguy'
author = 'Jean Mainguy'
release = __version__


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    # "sphinxcontrib.jquery",
    "sphinx.ext.duration",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.autodoc",
    'sphinx_search.extension'
]


source_suffix = {
    '.md': 'markdown'
}


templates_path = ['_templates']


# Prefix document path to section labels, to use:
# `path/to/file:heading` instead of just `heading`
autosectionlabel_prefix_document = True

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme' #'alabaster' # 

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']






"""
Unit tests for binette.

"""

import pytest

from binette import bin_manager
import networkx as nx

def test_get_all_possible_combinations():
    input_list = ["2", "3", "4"]
    expected_list = [("2", "3"), ("2", "4"), ("3", "4"), ("2", "3", "4")]

    assert list(bin_manager.get_all_possible_combinations(input_list)) == expected_list


@pytest.fixture
def example_bin_set1():
    bin1 = bin_manager.Bin(contigs={"1", "2"}, origin="test1", name="bin1")
    bin2 = bin_manager.Bin(contigs={"3", "4"}, origin="test1", name="bin2")
    bin3 = bin_manager.Bin(contigs={"5"}, origin="test1", name="bin2")
    return {bin1, bin2, bin3}
@pytest.fixture
def example_bin_set2():
    bin1 = bin_manager.Bin(contigs={"1", "2", "3"}, origin="test2", name="binA")
    bin2 = bin_manager.Bin(contigs={"4", "5"}, origin="test2", name="binB")
    return {bin1, bin2}

def test_bin_eq_true():
    bin1 = bin_manager.Bin(contigs={"1", "2", "e"}, origin="test1", name="bin1")
    bin2 = bin_manager.Bin(contigs={"1", "e", "2"}, origin="test2", name="binA")

    assert bin1 == bin2


def test_bin_eq_false():
    bin1 = bin_manager.Bin(contigs={"1", "2", "e"}, origin="test1", name="bin1")
    bin2 = bin_manager.Bin(contigs={"1", "e", "2", "33"}, origin="test2", name="binA")

    assert bin1 != bin2


def test_in_for_bin_list():
    bin1 = bin_manager.Bin(contigs={"1", "2", "e"}, origin="test1", name="bin1")
    bin2 = bin_manager.Bin(contigs={"1", "e", "2", "33"}, origin="test2", name="binA")
    bin3 = bin_manager.Bin(contigs={"4", "R"}, origin="test2", name="binA")

    bins = [bin1, bin2]

    assert bin1 in bins
    assert bin2 in bins
    assert bin3 not in bins

def test_add_length_positive_integer():
    bin_obj = bin_manager.Bin(contigs={"1", "2", "e"}, origin="test1", name="bin1")
    length = 100
    bin_obj.add_length(length)
    assert bin_obj.length == length

def test_add_length_negative_integer():
    bin_obj = bin_manager.Bin(contigs={"1", "2", "e"}, origin="test1", name="bin1") 
    with pytest.raises(ValueError):
        length = -50
        bin_obj.add_length(length)

def test_add_n50_positive_integer():
    bin_obj = bin_manager.Bin(contigs={"1", "2", "e"}, origin="test1", name="bin1")
    n50 = 100
    bin_obj.add_N50(n50)
    assert bin_obj.N50 == n50

def test_add_n50_negative_integer():
    bin_obj = bin_manager.Bin(contigs={"1", "2", "e"}, origin="test1", name="bin1") 
    with pytest.raises(ValueError):
        n50 = -50
        bin_obj.add_N50(n50)

def test_add_quality():

    completeness = 10
    contamination = 6
    contamination_weight = 2

    bin_obj = bin_manager.Bin(contigs={"1", "2", "e"}, origin="test1", name="bin1") 

    bin_obj.add_quality(completeness, contamination, contamination_weight)
    
    assert bin_obj.completeness == completeness 
    assert bin_obj.contamination == contamination 

    assert bin_obj.score == completeness - contamination * contamination_weight



    


def test_two_bin_intersection():
    bin1 = bin_manager.Bin(contigs={"1", "2", "e", "987"}, origin="test1", name="bin1")
    bin2 = bin_manager.Bin(contigs={"1", "e", "2", "33"}, origin="test2", name="binA")

    bin_intersection = bin1 & bin2

    assert bin_intersection == bin_manager.Bin({"1", "2", "e"}, "", "")


def test_multiple_bins_intersection():
    bin1 = bin_manager.Bin(contigs={"1", "2", "e", "987"}, origin="test1", name="bin1")
    bin2 = bin_manager.Bin(contigs={"1", "e", "2", "33"}, origin="test2", name="binA")
    bin3 = bin_manager.Bin(contigs={"1", "123", "2", "33"}, origin="test2", name="binA")

    bin_intersection = bin1.intersection(bin2, bin3)

    assert bin_intersection == bin_manager.Bin({"1", "2"}, "", "")


def test_bin_set_has_no_duplicate(example_bin_set1):
    set1_twice = list(example_bin_set1) + list(example_bin_set1)

    assert example_bin_set1 == set(set1_twice)


def test_bin_overlap_true():
    bin1 = bin_manager.Bin(contigs={"1", "2", "e"}, origin="test1", name="bin1")
    bin2 = bin_manager.Bin(contigs={"1", "e", "2", "33"}, origin="test2", name="binA")

    assert bin1.overlaps_with(bin2)

    assert bin2.overlaps_with(bin1)


def test_bin_overlap_false():
    bin1 = bin_manager.Bin(contigs={"13", "21", "ef"}, origin="test1", name="bin1")
    bin2 = bin_manager.Bin(contigs={"1", "e", "2", "33"}, origin="test2", name="binA")

    assert not bin1.overlaps_with(bin2)


def test_bin_union():
    bin1 = bin_manager.Bin(contigs={"13", "21"}, origin="test1", name="bin1")
    bin2 = bin_manager.Bin(contigs={"1", "e", "2", "33"}, origin="test2", name="binA")

    union_bin = bin_manager.Bin(contigs={"13", "21", "1", "e", "2", "33"}, origin="", name="")

    assert bin1.union(bin2) == union_bin
    assert bin1.union(bin2).name == "bin1 | binA"


def test_bin_union2():
    # Create some example bins
    bin1 = bin_manager.Bin({'contig1', 'contig2'}, 'origin1', 'bin1')
    bin2 = bin_manager.Bin({'contig2', 'contig3'}, 'origin2', 'bin2')
    bin3 = bin_manager.Bin({'contig4', 'contig5'}, 'origin3', 'bin3')

    # Perform union operation
    union_bin = bin1.union(bin2, bin3)

    # Check the result
    expected_contigs = {'contig1', 'contig2', 'contig3', 'contig4', 'contig5'}
    expected_name = 'bin1 | bin2 | bin3'
    expected_origin = 'union'

    assert union_bin.contigs == expected_contigs
    assert union_bin.name == expected_name
    assert union_bin.origin == expected_origin


def test_bin_difference():
    bin1 = bin_manager.Bin(contigs={5, 1, 6, 7, 8}, origin="test1", name="bin1")
    bin2 = bin_manager.Bin(contigs={3, 6, 7}, origin="test2", name="bin2")
    bin3 = bin_manager.Bin(contigs={1, 2, 3, 4}, origin="test2", name="bin3")

    diff_bin1_23 = bin_manager.Bin(contigs={5, 8}, origin="", name="")
    diff_bin1_2 = bin_manager.Bin(contigs={5, 1, 8}, origin="", name="")

    assert bin1.difference(bin2, bin3) == diff_bin1_23
    assert bin1.difference(bin2) == diff_bin1_2
    assert bin1.difference(bin2, bin3).name == "bin1 - bin2 - bin3"


def test_bin_intersection():
    bin1 = bin_manager.Bin(contigs={5, 1, 6, 7, 8}, origin="test1", name="bin1")
    bin2 = bin_manager.Bin(contigs={3, 6, 7}, origin="test2", name="bin2")
    bin3 = bin_manager.Bin(contigs={1, 2, 3, 4, 7}, origin="test2", name="bin3")

    inter_bin123 = bin_manager.Bin(contigs={7}, origin="", name="")
    iner_bin1_2 = bin_manager.Bin(contigs={7, 6}, origin="", name="")

    assert bin1.intersection(bin2, bin3) == inter_bin123
    assert bin1.intersection(bin2) == iner_bin1_2
    assert bin1.intersection(bin2, bin3).name == "bin1 & bin2 & bin3"


def test_select_best_bins_simple():

    b1 = bin_manager.Bin(contigs={1, 2}, origin="", name="")
    b2 = bin_manager.Bin(contigs={2}, origin="", name="")
    b3 = bin_manager.Bin(contigs={3}, origin="", name="")

    b1.score = 90
    b2.score = 80
    b3.score = 70

    b1.N50 = 100
    b2.N50 = 100
    b3.N50 = 100

    assert bin_manager.select_best_bins({b1, b2, b3}) == [b1, b3]


def test_select_best_bins_with_same_score():

    b1 = bin_manager.Bin(contigs={1, 2}, origin="", name="")
    b2 = bin_manager.Bin(contigs={2}, origin="", name="")
    b3 = bin_manager.Bin(contigs={3}, origin="", name="")

    b1.score = 90
    b2.score = 90
    b3.score = 70

    b1.N50 = 100
    b2.N50 = 101
    b3.N50 = 100

    assert bin_manager.select_best_bins({b1, b2, b3}) == [b2, b3]


def test_select_best_bins_with_equality():

    b1 = bin_manager.Bin(contigs={1, 2}, origin="", name="")
    b2 = bin_manager.Bin(contigs={2}, origin="", name="")
    b3 = bin_manager.Bin(contigs={3}, origin="", name="")

    b1.score = 90
    b2.score = 90
    b3.score = 70

    b1.N50 = 100
    b2.N50 = 100
    b3.N50 = 100

    # when score and n50 is the same, selection is made on the smallest id.
    # bin created first have a smaller id. so b1 should selected
    assert bin_manager.select_best_bins({b1, b2, b3}) == [b1, b3]


# The function should create intersection bins when there are overlapping contigs between bins.
def test_intersection_bins_created():
    set1 = [
        bin_manager.Bin(contigs={"1", "2"}, origin="A", name="bin1"),
        bin_manager.Bin(contigs={"3", "4"}, origin="A", name="bin2"),
        bin_manager.Bin(contigs={"5"}, origin="A", name="bin2"),
    ]
    # need to defined completeness and conta
    # because when too low the bin is not used in all operation
    for b in set1:
        b.completeness = 100
        b.contamination = 0

    binA = bin_manager.Bin(contigs={"1", "3"}, origin="B", name="binA")
    binA.contamination = 0
    binA.completeness = 100
    set2 = [
        binA,
    ]
    bin_set_name_to_bins = {'set1': set1,
                            "set2":set2}

    intermediate_bins_result = bin_manager.create_intermediate_bins(bin_set_name_to_bins)

    expected_intermediate_bins = {bin_manager.Bin(contigs={"1", "2", "3"}, origin="bin1 | binA ", name="NA"),
                                  bin_manager.Bin(contigs={"2"}, origin="bin1 - binA ", name="NA"),
                                  bin_manager.Bin(contigs={"1"}, origin="bin1 & binA ", name="NA"),
                                  bin_manager.Bin(contigs={"1", "4", "3"}, origin="bin2 | binA ", name="NA"),
                                  bin_manager.Bin(contigs={"4"}, origin="bin2 - binA ", name="NA"), # binA -bin1 is equal to bin1 & binA
                                  bin_manager.Bin(contigs={"3"}, origin="bin1 & binA ", name="NA"),
                                  }

    assert intermediate_bins_result == expected_intermediate_bins


# Renames contigs in bins based on provided mapping.
def test_renames_contigs(example_bin_set1):
    
    bin_set = [
        bin_manager.Bin(contigs={"c1", "c2"}, origin="A", name="bin1"),
        bin_manager.Bin(contigs={"c3", "c4"}, origin="A", name="bin2")
    ]
        
    contig_to_index = {'c1': 1, 'c2': 2, 'c3': 3, 'c4': 4, "c5":5}

    # Act
    bin_manager.rename_bin_contigs(bin_set, contig_to_index)

    # Assert
    assert bin_set[0].contigs == {1, 2}
    assert bin_set[0].hash == hash(str(sorted({1, 2})))
    assert bin_set[1].contigs == {3, 4}
    assert bin_set[1].hash == hash(str(sorted({3, 4})))

def test_get_contigs_in_bins():
    bin_set = [
        bin_manager.Bin(contigs={"c1", "c2"}, origin="A", name="bin1"),
        bin_manager.Bin(contigs={"c3", "c4"}, origin="A", name="bin2"),
        bin_manager.Bin(contigs={"c3", "c18"}, origin="A", name="bin2")
    ]

    contigs = bin_manager.get_contigs_in_bins(bin_set)

    assert contigs == {"c1", "c2", "c3", "c4", "c18"}


def test_dereplicate_bin_sets():
    b1 = bin_manager.Bin(contigs={"c1", "c2"}, origin="A", name="bin1")
    b2 = bin_manager.Bin(contigs={"c3", "c4"}, origin="A", name="bin2")
    b3 = bin_manager.Bin(contigs={"c3", "c18"}, origin="A", name="bin2")

    bdup  = bin_manager.Bin(contigs={"c3", "c18"}, origin="D", name="C")

    derep_bins_result =  bin_manager.dereplicate_bin_sets([[b2,b3 ],[b1, bdup]])


    assert derep_bins_result == {b1, b2, b3}



def test_from_bin_sets_to_bin_graph():


    bin1 = bin_manager.Bin(contigs={"1", "2"}, origin="A", name="bin1")
    bin2 = bin_manager.Bin(contigs={"3", "4"}, origin="A", name="bin2")
    bin3 = bin_manager.Bin(contigs={"5"}, origin="A", name="bin3")

    set1 = [bin1,bin2,bin3]

    binA = bin_manager.Bin(contigs={"1", "3"}, origin="B", name="binA")
            
    set2 = [binA]

    result_graph = bin_manager.from_bin_sets_to_bin_graph({"B":set2, "A":set1}) 

    assert result_graph.number_of_edges() == 2
    # bin3 is not connected to any bin so it is not in the graph
    assert result_graph.number_of_nodes() == 3

    assert set(result_graph.nodes) == {binA, bin1, bin2}

@pytest.fixture
def simple_bin_graph():

    bin1 = bin_manager.Bin(contigs={"1", "2", "3"}, origin="A", name="bin1")
    bin2 = bin_manager.Bin(contigs={"1", "2", "4"}, origin="B", name="bin2")

    for b in [bin1, bin2]:
        b.completeness = 100
        b.contamination = 0
    
    G = nx.Graph()
    G.add_edge(bin1, bin2)

    return G


def test_get_intersection_bins(simple_bin_graph):

    intersec_bins = bin_manager.get_intersection_bins(simple_bin_graph)
    
    assert len(intersec_bins) == 1
    intersec_bin = intersec_bins.pop()

    assert intersec_bin.contigs == {"1", "2"}

def test_get_difference_bins(simple_bin_graph):
    
    difference_bins = bin_manager.get_difference_bins(simple_bin_graph)
    
    expected_bin1 = bin_manager.Bin(contigs={"3"}, origin="D", name="1")
    expected_bin2 = bin_manager.Bin(contigs={"4"}, origin="D", name="2")

    assert len(difference_bins) == 2
    assert difference_bins == {expected_bin1,expected_bin2}


def test_get_union_bins(simple_bin_graph):
    
    u_bins = bin_manager.get_union_bins(simple_bin_graph)
    
    expected_bin1 = bin_manager.Bin(contigs={"1", "2", "3", "4"}, origin="U", name="1")

    assert len(u_bins) == 1
    assert u_bins == {expected_bin1}




def test_get_bins_from_contig2bin_table(tmp_path):
    # Create a temporary file (contig-to-bin table) for testing
    test_table_content = [
        "# Sample contig-to-bin table",
        "contig1\tbin1",
        "contig2\tbin1",
        "contig3\tbin2",
    ]
    test_table_path = tmp_path / "test_contig2bin_table.txt"
    test_table_path.write_text("\n".join(test_table_content))

    # Define set name for the bins
    set_name = "TestSet"

    # Call the function to generate Bin objects
    result_bins = bin_manager.get_bins_from_contig2bin_table(str(test_table_path), set_name)

    # Validate the result
    assert len(result_bins) == 2  # Check if the correct number of bins are created

    # Define expected bins based on the test table content
    expected_bins = [
            bin_manager.Bin(contigs={"contig1", "contig2"}, origin="A", name="bin1"),
            bin_manager.Bin(contigs={"contig3"}, origin="A", name="bin2")
    ]

    # Compare expected bins with the result
    assert all(expected_bin in result_bins for expected_bin in expected_bins)
    assert all(result_bin in expected_bins for result_bin in result_bins)


def test_parse_contig2bin_tables(tmp_path):
    # Create temporary contig-to-bin tables for testing
    test_tables = {
        "set1": [
            "# Sample contig-to-bin table for bin1",
            "contig1\tbin1",
            "contig2\tbin1",
            "contig3\tbin2",
        ],
        "set2": [
            "# Sample contig-to-bin table for bin2",
            "contig3\tbinA",
            "contig4\tbinA"
        ]
    }

    # Create temporary files for contig-to-bin tables
    for name, content in test_tables.items():
        table_path = tmp_path / f"test_{name}_contig2bin_table.txt"
        table_path.write_text("\n".join(content))

    # Call the function to parse contig-to-bin tables
    result_bin_dict = bin_manager.parse_contig2bin_tables({name: str(tmp_path / f"test_{name}_contig2bin_table.txt") for name in test_tables})

    # Validate the result
    assert len(result_bin_dict) == len(test_tables)  # Check if the number of bins matches the number of tables

    # Define expected Bin objects based on the test tables
    expected_bins = {
        "set1": [
            bin_manager.Bin(contigs={"contig1", "contig2"}, origin="set1", name="bin1"),
            bin_manager.Bin(contigs={"contig3"}, origin="set1", name="bin2"),
        ],
        "set2": [
            bin_manager.Bin(contigs={"contig3", "contig4"}, origin="set2", name="binA"),
        ]
    }

    # Compare expected bins with the result
    for name, expected in expected_bins.items():
        assert name in result_bin_dict
        assert len(result_bin_dict[name]) == len(expected)
        for result_bin, expected_bin in zip(result_bin_dict[name], expected):
            assert result_bin.contigs == expected_bin.contigs
            assert result_bin.name == expected_bin.name
            assert result_bin.origin == expected_bin.origin




@pytest.fixture
def create_temp_bin_files(tmpdir):
    # Create temporary bin files
    bin_dir = tmpdir.mkdir("bins")
    bin1 = bin_dir.join("bin1.fasta")
    bin1.write(">contig1\nATGC\n>contig2\nGCTA")

    bin2 = bin_dir.join("bin2.fasta")
    bin2.write(">contig3\nTTAG\n>contig4\nCGAT")

    return bin_dir

@pytest.fixture
def create_temp_bin_directories(tmpdir, create_temp_bin_files):
    # Create temporary bin directories
    bin_dir1 = tmpdir.mkdir("set1")
    bin1 = bin_dir1.join("bin1.fasta")
    bin1.write(">contig1\nATGC\n>contig2\nGCTA")

    bin2 = bin_dir1.join("bin2.fasta")
    bin2.write(">contig3\nTTAG\n>contig4\nCGAT")


    bin_dir2 = tmpdir.mkdir("set2")
    bin2 = bin_dir2.join("binA.fasta")
    bin2.write(">contig3\nTTAG\n>contig4\nCGAT\n>contig5\nCGGC")

    return {"set1": str(bin_dir1), "set2": str(bin_dir2)}


def test_get_bins_from_directory(create_temp_bin_files):
    bin_dir = create_temp_bin_files
    set_name = "TestSet"

    bins = bin_manager.get_bins_from_directory(str(bin_dir), set_name)

    assert len(bins) == 2  # Ensure that the correct number of Bin objects is returned

    # Check if the Bin objects are created with the correct contigs, set name, and bin names
    assert isinstance(bins[0], bin_manager.Bin)
    assert isinstance(bins[1], bin_manager.Bin)
    assert bins[1].contigs in [{"contig1", "contig2"}, {"contig3", "contig4"}]
    assert bins[0].contigs in [{"contig1", "contig2"}, {"contig3", "contig4"}]
    assert bins[0].origin == set_name
    assert bins[1].origin == set_name
    assert bins[1].name in ["bin2.fasta", "bin1.fasta"]
    assert bins[0].name in ["bin2.fasta", "bin1.fasta"]

def test_get_bins_from_directory_no_files(tmpdir):
    bin_dir = str(tmpdir.mkdir("empty_bins"))
    set_name = "EmptySet"

    bins = bin_manager.get_bins_from_directory(bin_dir, set_name)

    assert len(bins) == 0  # Ensure that no Bin objects are returned for an empty directory




def test_parse_bin_directories(create_temp_bin_directories):
    set_name_to_bin_dir = create_temp_bin_directories

    bins = bin_manager.parse_bin_directories(set_name_to_bin_dir)

    assert len(bins) == 2  # Ensure that the correct number of bin directories is parsed

    # Check if the Bin objects are created with the correct contigs, set name, and bin names
    assert isinstance(bins["set1"][0], bin_manager.Bin)
    assert isinstance(bins["set2"][0], bin_manager.Bin)

    assert len(bins["set2"]) == 1
    assert len(bins["set1"]) == 2





import logging
import os
import pyfastx
from typing import List, Dict
import csv

from binette.bin_manager import Bin


def infer_bin_name_from_bin_inputs(input_bins: List[str]) -> Dict[str, str]:
    """
    Infer bin names from a list of bin input directories.

    :param input_bins: List of input bin directories.
    :return: Dictionary mapping inferred bin names to their corresponding directories.
    """
    logging.debug(f"Inferring bin names from input bins:")

    commonprefix_len = len(os.path.commonprefix(input_bins))
    reversed_strings = [s[::-1] for s in input_bins]
    commonsufix_len = len(os.path.commonprefix(reversed_strings))

    bin_name_to_bin_dir = {d[commonprefix_len: len(d) - commonsufix_len]: d for d in input_bins}

    logging.debug(f"Input bins:  {' '.join(input_bins)}")
    logging.debug(f"Common prefix to remove: {os.path.commonprefix(reversed_strings)[::-1]}")
    logging.debug(f"Common suffix to remove: {os.path.commonprefix(input_bins)}")
    logging.debug(f"bin_name_to_bin_dir: {bin_name_to_bin_dir}")

    return bin_name_to_bin_dir


def write_bin_info(bins: List[Bin], output: str, add_contigs: bool = False):
    """
    Write bin information to a TSV file.

    :param bins: List of Bin objects.
    :param output: Output file path for writing the TSV.
    :param add_contigs: Flag indicating whether to include contig information.
    """

    header = ["bin_id", "origin", "name", "completeness", "contamination", "score", "size", "N50", "contig_count"]
    if add_contigs:
        header.append('contigs')

    bin_infos = []
    for bin_obj in sorted(bins, key=lambda x: (x.score, x.N50, -x.id), reverse=True):
        bin_info = [
            bin_obj.id,
            bin_obj.origin,
            bin_obj.name,
            bin_obj.completeness,
            bin_obj.contamination,
            bin_obj.score,
            bin_obj.length,
            bin_obj.N50,
            len(bin_obj.contigs),
        ]
        if add_contigs:
            bin_info.append(";".join(str(c) for c in bin_obj.contigs) if add_contigs else "")

        bin_infos.append(bin_info)

    with open(output, "w", newline="") as fl:
        writer = csv.writer(fl, delimiter="\t")
        writer.writerow(header)
        writer.writerows(bin_infos)


def write_bins_fasta(selected_bins: List[Bin], contigs_fasta: str, outdir: str):
    """
    Write selected bins' contigs to separate FASTA files.

    :param selected_bins: List of Bin objects representing the selected bins.
    :param contigs_fasta: Path to the input FASTA file containing contig sequences.
    :param outdir: Output directory to save the individual bin FASTA files.
    """

    fa = pyfastx.Fasta(contigs_fasta, build_index=True)

    for sbin in selected_bins:
        outfile = os.path.join(outdir, f"bin_{sbin.id}.fa")

        with open(outfile, "w") as outfl:
            sequences = (f">{c}\n{fa[c]}" for c in sbin.contigs)
            outfl.write("\n".join(sequences) + "\n")


def check_contig_consistency(contigs_from_assembly: List[str],
                             contigs_from_elsewhere: List[str],
                             assembly_file: str,
                             elsewhere_file: str ):
    """
    Check the consistency of contig names between different sources.

    :param contigs_from_assembly: List of contig names from the assembly file.
    :param contigs_from_elsewhere: List of contig names from an external source.
    :param assembly_file: Path to the assembly file.
    :param elsewhere_file: Path to the file from an external source.
    :raises AssertionError: If inconsistencies in contig names are found.
    """
    logging.debug("check_contig_consistency.")
    are_contigs_consistent = len(set(contigs_from_elsewhere) | set(contigs_from_assembly)) <= len(
        set(contigs_from_assembly)
    )

    issue_countigs = len(set(contigs_from_elsewhere) - set(contigs_from_assembly))
    
    message = f"{issue_countigs} contigs found in file {elsewhere_file} \
                were not found in assembly_file ({assembly_file})."
    assert are_contigs_consistent, message


def check_resume_file(faa_file: str, diamond_result_file: str) -> None:
    """
    Check the existence of files required for resuming the process.

    :param faa_file: Path to the protein file.
    :param diamond_result_file: Path to the Diamond result file.
    :raises FileNotFoundError: If the required files don't exist for resuming.
    """

    if os.path.isfile(faa_file) and os.path.isfile(diamond_result_file):
        return

    if not os.path.isfile(faa_file):
        error_msg = f"Protein file '{faa_file}' does not exist. Resuming is not possible."
        logging.error(error_msg)
        raise FileNotFoundError(error_msg)

    if not os.path.isfile(diamond_result_file):
        error_msg = f"Diamond result file '{diamond_result_file}' does not exist. Resuming is not possible."
        logging.error(error_msg)
        raise FileNotFoundError(error_msg)





from binette import cds
import pytest
import pyrodigal

from pathlib import Path
from unittest.mock import mock_open, patch

class MockContig:
    def __init__(self, name, seq):
        self.seq = seq
        self.name = name

@pytest.fixture
def contig1():
    contig =  MockContig(name="contig1", 
                          seq="ATGAGCATCAGGGGAGTAGGAGGATGCAACGGGAATAGTCGAATCCCTTCTCATAATGGGGATGGATCGAATCGCAGAAGTCAAAATACGAAGGGTAATAATAAAGTTGAAGATCGAGTTTGT")
    return contig

@pytest.fixture
def contig2():
    contig =  MockContig(name="contig2", 
                          seq="TTGGTCGTATGACTGATAATTTCTCAGACATTGAAAACTTTAATGAAATTTTCAACAGAAAACCTGCTTTACAATTTCGTTTTTA")
    return contig

@pytest.fixture
def orf_finder():
    try:
        # for version >=3 of pyrodigal
        orf_finder = pyrodigal.GeneFinder(meta="meta")
    except AttributeError:
        orf_finder = pyrodigal.OrfFinder(meta="meta")

    return orf_finder

# Predict open reading frames with Pyrodigal using 1 thread.
def test_predict_orf_with_1_thread(contig1, contig2):

    contigs_iterator = [contig1, contig2]
    outfaa = "output.fasta"
    threads = 1

    result = cds.predict(contigs_iterator, outfaa, threads)

    assert isinstance(result, dict)
    assert len(result) == 2
    assert "contig1" in result
    assert "contig2" in result
    assert isinstance(result["contig1"], list)
    assert isinstance(result["contig2"], list)
    assert len(result["contig1"]) == 1
    assert len(result["contig2"]) == 1
    assert isinstance(result["contig1"][0], str)
    assert isinstance(result["contig2"][0], str)


def test_predict_orf_with_multiple_threads(contig1, contig2):

    contigs_iterator = [contig1, contig2]
    outfaa = "output.fasta"
    threads = 4

    result = cds.predict(contigs_iterator, outfaa, threads)
    
    assert isinstance(result, dict)
    assert len(result) == 2
    assert "contig1" in result
    assert "contig2" in result
    assert isinstance(result["contig1"], list)
    assert isinstance(result["contig2"], list)
    assert len(result["contig1"]) == 1
    assert len(result["contig2"]) == 1
    assert isinstance(result["contig1"][0], str)
    assert isinstance(result["contig2"][0], str)


def test_predict_genes(contig1, orf_finder):


    result = cds.predict_genes(orf_finder.find_genes, contig1)

    assert isinstance(result, tuple)
    assert len(result) == 2
    assert isinstance(result[0], str)
    assert result[0] == contig1.name
    assert result[0] == "contig1"


# Extract the contig name from a CDS name.
def test_extract_contig_name_from_cds_name():
    cds_name = "contig1_gene1"

    result = cds.get_contig_from_cds_name(cds_name)

    assert isinstance(result, str)
    assert result == "contig1"

def test_extract_contig_name_from_cds_name():
    cds_name = "contig1_gene1"

    result = cds.get_contig_from_cds_name(cds_name)

    assert isinstance(result, str)
    assert result == "contig1"


# Import the functions write_faa and parse_faa_file here


def test_write_faa(contig1, orf_finder):
    
    predicted_genes = orf_finder.find_genes(contig1.seq)
    contig_name = 'contig'
    output_file = "tests/tmp_file.faa"

    cds.write_faa(output_file, [(contig_name, predicted_genes)])

    # Check if the file was created and first seq starts
    # with the contig name as expected
    assert Path(output_file).exists()
    with open(output_file, "r") as f:
        assert f.read().startswith(f">{contig_name}")


def test_parse_faa_file(tmp_path):
    # Mock a FASTA file
    fasta_content = (
        ">contig1_gene1\n"
        "AAAAAAAAAAA\n"
        ">contig1_gene2\n"
        "CCCCCCCCCCC\n"
        ">contig2_gene1\n"
        "TTTTTTTTTTTT\n"
    )
    faa_file = tmp_path / "mock_file.faa"
    faa_file.write_text(fasta_content)

    # Call the function
    result = cds.parse_faa_file(faa_file)

    # Check if the output matches the expected dictionary
    expected_result = {
        'contig1': ['AAAAAAAAAAA', 'CCCCCCCCCCC'],
        'contig2': ['TTTTTTTTTTTT']
    }
    assert result == expected_result

def test_get_aa_composition():

    genes = ['AAAA',
            "CCCC",
            "TTTT",
            "GGGG"]

    result = cds.get_aa_composition(genes)

    assert dict(result) == {'A': 4, 'C': 4, 'T': 4, 'G': 4}

def test_get_contig_cds_metadata_flat():

    contig_to_genes = {"c1":["AAAA", "GGGG", "CCCC"],
                       "c2":["TTTT", "CCCC"]}

    contig_to_cds_count, contig_to_aa_counter, contig_to_aa_length = cds.get_contig_cds_metadata_flat(contig_to_genes)
    
    assert contig_to_cds_count == {"c1":3, "c2":2}
    assert contig_to_aa_counter == {"c1": {'A': 4, 'G': 4, "C":4} , "c2":{'C': 4, 'T': 4}}
    assert contig_to_aa_length == {"c1":12, "c2":8}

def test_get_contig_cds_metadata():

    contig_to_genes = {"c1":["AAAA", "GGGG", "CCCC"],
                       "c2":["TTTT", "CCCC"]}

    contig_metadata = cds.get_contig_cds_metadata(contig_to_genes, 1)
    
    assert contig_metadata['contig_to_cds_count'] == {"c1":3, "c2":2}
    assert contig_metadata['contig_to_aa_counter'] == {"c1": {'A': 4, 'G': 4, "C":4} , "c2":{'C': 4, 'T': 4}}
    assert contig_metadata['contig_to_aa_length'] == {"c1":12, "c2":8}


#!/usr/bin/env python
"""
Module      : Main
Description : The main entry point for the program.
Copyright   : (c) Jean Mainguy, 28 nov. 2022 
License     : MIT
Maintainer  : Jean Mainguy
Portability : POSIX
"""

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

import sys
import logging
import os

import binette
from binette import contig_manager, cds, diamond, bin_quality, bin_manager, io_manager as io
from typing import List, Dict, Set, Tuple


def init_logging(verbose, debug):
    """Initialise logging."""
    if debug:
        level = logging.DEBUG
    elif verbose:
        level = logging.INFO
    else:
        level = logging.WARNING

    logging.basicConfig(
        level=level,
        format="%(asctime)s %(levelname)s - %(message)s",
        datefmt="[%Y-%m-%d %H:%M:%S]",
    )

    logging.info("Program started")
    logging.info(
        f'command line: {" ".join(sys.argv)}',
    )



def parse_arguments(args):
    """Parse script arguments."""

    parser = ArgumentParser(
        description=f"Binette version={binette.__version__}",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )

    # Input arguments category
    input_group = parser.add_argument_group('Input Arguments')
    input_arg = input_group.add_mutually_exclusive_group(required=True)

    input_arg.add_argument(
        "-d",
        "--bin_dirs",
        nargs="+",
        help="List of bin folders containing each bin in a fasta file.",
    )

    input_arg.add_argument(
        "-b",
        "--contig2bin_tables",
        nargs="+",
        help="List of contig2bin table with two columns separated\
            with a tabulation: contig, bin",
    )

    input_group.add_argument("-c", "--contigs", required=True, help="Contigs in fasta format.")

    # Other parameters category
    other_group = parser.add_argument_group('Other Arguments')

    other_group.add_argument(
        "-m",
        "--min_completeness",
        default=40,
        type=int,
        help="Minimum completeness required for final bin selections.",
    )

    other_group.add_argument("-t", "--threads", default=1, type=int, help="Number of threads to use.")

    other_group.add_argument("-o", "--outdir", default="results", help="Output directory.")

    other_group.add_argument(
        "-w",
        "--contamination_weight",
        default=2,
        type=float,
        help="Bin are scored as follow: completeness - weight * contamination. "
             "A low contamination_weight favor complete bins over low contaminated bins.",
    )

    other_group.add_argument(
        "--checkm2_db",
        help="Provide a path for the CheckM2 diamond database. "
        "By default the database set via <checkm2 database> is used.",
    )

    other_group.add_argument("--low_mem", help="Use low mem mode when running diamond", action="store_true")

    other_group.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")

    other_group.add_argument("--debug", help="Activate debug mode", action="store_true")

    other_group.add_argument("--resume",
                            action="store_true",
                            help="Activate resume mode. Binette will examine the 'temporary_files' directory "
                                "within the output directory and reuse any existing files if possible."
                        )


    other_group.add_argument("--version", action="version", version=binette.__version__)

    args = parser.parse_args(args)
    return args

def parse_input_files(bin_dirs: List[str], contig2bin_tables: List[str], contigs_fasta: str) -> Tuple[Dict[str, List], List, Dict[str, List], Dict[str, int]]:
    """
    Parses input files to retrieve information related to bins and contigs.

    :param bin_dirs: List of paths to directories containing bin FASTA files.
    :param contig2bin_tables: List of paths to contig-to-bin tables.
    :param contigs_fasta: Path to the contigs FASTA file.

    :return: A tuple containing:
        - Dictionary mapping bin set names to lists of bins.
        - List of original bins.
        - Dictionary mapping bins to lists of contigs.
        - Dictionary mapping contig names to their lengths.
    """

    if bin_dirs:
        logging.info("Parsing bin directories.")
        bin_name_to_bin_dir = io.infer_bin_name_from_bin_inputs(bin_dirs)
        bin_set_name_to_bins = bin_manager.parse_bin_directories(bin_name_to_bin_dir)
    else:
        logging.info("Parsing bin2contig files.")
        bin_name_to_bin_table = io.infer_bin_name_from_bin_inputs(contig2bin_tables)
        bin_set_name_to_bins = bin_manager.parse_contig2bin_tables(bin_name_to_bin_table)

    logging.info(f"{len(bin_set_name_to_bins)} bin sets processed:")
    for bin_set_id, bins in bin_set_name_to_bins.items():
        logging.info(f" {bin_set_id} - {len(bins)} bins")

    original_bins = bin_manager.dereplicate_bin_sets(bin_set_name_to_bins.values())
    contigs_in_bins = bin_manager.get_contigs_in_bins(original_bins)

    logging.info(f"Parsing contig fasta file: {contigs_fasta}")
    contigs_object = contig_manager.parse_fasta_file(contigs_fasta)
    contig_to_length = {seq.name: len(seq) for seq in contigs_object if seq.name in contigs_in_bins}

    return bin_set_name_to_bins, original_bins, contigs_in_bins, contig_to_length


def manage_protein_alignement(faa_file: str, contigs_fasta: str, contig_to_length: Dict[str, List],
                                contigs_in_bins: Dict[str, List], diamond_result_file: str,
                                checkm2_db: str, threads: int, resume: bool, low_mem: bool) -> Tuple[Dict[str, int], Dict[str, int]]:
    """
    Predicts or reuses proteins prediction and runs diamond on them.
    
    :param faa_file: The path to the .faa file.
    :param contigs_fasta: The path to the contigs FASTA file.
    :param contig_to_length: Dictionary mapping contig names to their lengths.
    :param contigs_in_bins: Dictionary mapping bin names to lists of contigs.
    :param diamond_result_file: The path to the diamond result file.
    :param checkm2_db: The path to the CheckM2 database.
    :param threads: Number of threads for parallel processing.
    :param resume: Boolean indicating whether to resume the process.
    :param low_mem: Boolean indicating whether to use low memory mode.

    :return: A tuple containing dictionaries - contig_to_kegg_counter and contig_to_genes.
    """

    # Predict or reuse proteins prediction and run diamond on them
    if resume:
        logging.info(f"Parsing faa file: {faa_file}.")
        contig_to_genes = cds.parse_faa_file(faa_file)
        io.check_contig_consistency(contig_to_length, contig_to_genes, contigs_fasta, faa_file)

    else:
        contigs_iterator = (s for s in contig_manager.parse_fasta_file(contigs_fasta) if s.name in contigs_in_bins)
        contig_to_genes = cds.predict(contigs_iterator, faa_file, threads)

        if checkm2_db:
            diamond_db_path = checkm2_db
        else:
            # get checkm2 db stored in checkm2 install
            diamond_db_path = diamond.get_checkm2_db()
        
        diamond_log = f"{os.path.splitext(diamond_result_file)[0]}.log"

        diamond.run(
            faa_file,
            diamond_result_file,
            diamond_db_path,
            diamond_log,
            threads,
            low_mem=low_mem,
        )

    logging.info("Parsing diamond results.")
    contig_to_kegg_counter = diamond.get_contig_to_kegg_id(diamond_result_file)

    # Check contigs from diamond vs input assembly consistency
    io.check_contig_consistency(contig_to_length, contig_to_kegg_counter, contigs_fasta, diamond_result_file)

    return contig_to_kegg_counter, contig_to_genes


def select_bins_and_write_them(all_bins: Set[bin_manager.Bin], contigs_fasta: str, final_bin_report: str, min_completeness: float,
                               index_to_contig: dict, outdir: str, debug: bool) -> List[bin_manager.Bin]:
    """
    Selects and writes bins based on specific criteria.

    :param all_bins: Set of Bin objects.
    :param contigs_fasta: Path to the contigs FASTA file.
    :param final_bin_report: Path to write the final bin report.
    :param min_completeness: Minimum completeness threshold for bin selection.
    :param index_to_contig: Dictionary mapping indices to contig names.
    :param outdir: Output directory to save final bins and reports.
    :param debug: Debug mode flag.
    :return: Selected bins that meet the completeness threshold.
    """

    outdir_final_bin_set = os.path.join(outdir, "final_bins")
    os.makedirs(outdir_final_bin_set, exist_ok=True)

    if debug:
        all_bins_for_debug = set(all_bins)
        all_bin_compo_file = os.path.join(outdir, "all_bins_quality_reports.tsv")
        
        logging.info(f"Writing all bins in {all_bin_compo_file}")
        
        io.write_bin_info(all_bins_for_debug, all_bin_compo_file, add_contigs=True)
        
        with open(os.path.join(outdir, "index_to_contig.tsv"), 'w') as flout:
            flout.write('\n'.join((f'{i}\t{c}' for i, c in index_to_contig.items())))

    logging.info("Selecting best bins")
    selected_bins = bin_manager.select_best_bins(all_bins)

    logging.info(f"Bin Selection: {len(selected_bins)} selected bins")

    logging.info(f"Filtering bins: only bins with completeness >= {min_completeness} are kept")
    selected_bins = [b for b in selected_bins if b.completeness >= min_completeness]

    logging.info(f"Filtering bins: {len(selected_bins)} selected bins")

    logging.info(f"Writing selected bins in {final_bin_report}")
    
    for b in selected_bins:
        b.contigs = {index_to_contig[c_index] for c_index in b.contigs}
    
    io.write_bin_info(selected_bins, final_bin_report)

    io.write_bins_fasta(selected_bins, contigs_fasta, outdir_final_bin_set)

    return selected_bins



def log_selected_bin_info(selected_bins: List[bin_manager.Bin], hq_min_completeness: float, hq_max_conta: float):
    """
    Log information about selected bins based on quality thresholds.

    :param selected_bins: List of Bin objects to analyze.
    :param hq_min_completeness: Minimum completeness threshold for high-quality bins.
    :param hq_max_conta: Maximum contamination threshold for high-quality bins.

    This function logs information about selected bins that meet specified quality thresholds.
    It counts the number of high-quality bins based on completeness and contamination values.
    """

    # Log completeness and contamination in debug log
    logging.debug("High quality bins:")
    for sb in selected_bins:
        if sb.completeness >= hq_min_completeness and sb.contamination <= hq_max_conta:
            logging.debug(f"> {sb} completeness={sb.completeness}, contamination={sb.contamination}")

    # Count high-quality bins and single-contig high-quality bins
    hq_bins = len([sb for sb in selected_bins if sb.completeness >= hq_min_completeness and sb.contamination <= hq_max_conta])

    # Log information about high-quality bins
    thresholds = f"(completeness >= {hq_min_completeness} and contamination <= {hq_max_conta})"
    logging.info(f"{hq_bins}/{len(selected_bins)} selected bins have a high quality {thresholds}.")


def main():
    "Orchestrate the execution of the program"

    args = parse_arguments(sys.argv[1:]) # sys.argv is passed in order to be able to test the function parse_arguments

    init_logging(args.verbose, args.debug)

    # High quality threshold used just to log number of high quality bins.
    hq_max_conta = 5
    hq_min_completeness = 90

    # Temporary files #
    out_tmp_dir = os.path.join(args.outdir, "temporary_files")
    os.makedirs(out_tmp_dir, exist_ok=True)

    faa_file = os.path.join(out_tmp_dir, "assembly_proteins.faa")
    diamond_result_file = os.path.join(out_tmp_dir, "diamond_result.tsv")

    # Output files #
    final_bin_report = os.path.join(args.outdir, "final_bins_quality_reports.tsv")


    if args.resume:
        io.check_resume_file(faa_file, diamond_result_file)

    bin_set_name_to_bins, original_bins, contigs_in_bins, contig_to_length = parse_input_files(args.bin_dirs, args.contig2bin_tables, args.contigs)

    contig_to_kegg_counter, contig_to_genes = manage_protein_alignement(faa_file=faa_file, contigs_fasta=args.contigs, contig_to_length=contig_to_length,
                                contigs_in_bins=contigs_in_bins,
                                diamond_result_file=diamond_result_file, checkm2_db=args.checkm2_db,
                                threads=args.threads, resume=args.resume, low_mem=args.low_mem)
    
    # Use contig index instead of contig name to save memory
    contig_to_index, index_to_contig = contig_manager.make_contig_index(contigs_in_bins)

    contig_to_kegg_counter = contig_manager.apply_contig_index(contig_to_index, contig_to_kegg_counter)
    contig_to_genes = contig_manager.apply_contig_index(contig_to_index, contig_to_genes)
    contig_to_length = contig_manager.apply_contig_index(contig_to_index, contig_to_length)

    bin_manager.rename_bin_contigs(original_bins, contig_to_index)


    # Extract cds metadata ##
    logging.info("Compute cds metadata.")
    contig_metadat = cds.get_contig_cds_metadata(contig_to_genes, args.threads)

    contig_metadat["contig_to_kegg_counter"] = contig_to_kegg_counter
    contig_metadat["contig_to_length"] = contig_to_length
    

    logging.info("Add size and assess quality of input bins")
    bin_quality.add_bin_metrics(original_bins, contig_metadat, args.contamination_weight, args.threads)

    logging.info("Create intermediate bins:")
    new_bins = bin_manager.create_intermediate_bins(bin_set_name_to_bins)

    logging.info("Assess quality for supplementary intermediate bins.")
    new_bins = bin_quality.add_bin_metrics(new_bins, contig_metadat, args.contamination_weight, args.threads)


    logging.info("Dereplicating input bins and new bins")
    all_bins = original_bins | new_bins

    selected_bins = select_bins_and_write_them(all_bins,  args.contigs, final_bin_report, args.min_completeness, index_to_contig,  args.outdir, args.debug)

    log_selected_bin_info(selected_bins, hq_min_completeness, hq_max_conta)

    return 0


import pytest
from binette import io_manager
from pathlib import Path





class Bin:
    def __init__(self, bin_id, origin, name, completeness, contamination, score, length, N50, contigs):
        self.id = bin_id
        self.origin = origin
        self.name = name
        self.completeness = completeness
        self.contamination = contamination
        self.score = score
        self.length = length
        self.N50 = N50
        self.contigs = contigs

@pytest.fixture
def bin1():
    return Bin(1, 'origin1', 'name1', 90, 5, 80, 1000, 500, ['contig1', 'contig3'])

@pytest.fixture
def bin2():
    return Bin(2, 'origin2', 'name2', 85, 8, 75, 1200, 600, ['contig2', 'contig4'])


def test_infer_bin_name_from_bin_inputs():
    # Mock input data
    input_bins = [
        '/path/to/bin1',
        '/path/to/bin2',
        '/path/to/bin3'
    ]

    # Call the function
    result = io_manager.infer_bin_name_from_bin_inputs(input_bins)

    # Define the expected output
    expected_result = {
        '1': '/path/to/bin1',
        '2': '/path/to/bin2',
        '3': '/path/to/bin3'
    }

    # Check if the output matches the expected dictionary
    assert result == expected_result


def test_write_bin_info(tmp_path, bin1, bin2):
    # Mock input data
    bins = [bin1, bin2]

    output_file = tmp_path / "output.tsv"

    # Call the function
    io_manager.write_bin_info(bins, output_file)

    # Check if the file was created and its content matches the expected output
    assert Path(output_file).exists()

    with open(output_file, "r") as f:
        content = f.read()
        assert "bin_id\torigin\tname\tcompleteness\tcontamination\tscore\tsize\tN50\tcontig_count" in content
        assert "1\torigin1\tname1\t90\t5\t80\t1000\t500\t2" in content
        assert "2\torigin2\tname2\t85\t8\t75\t1200\t600\t2" in content


def test_write_bin_info_add_contig(tmp_path, bin1, bin2):
    # Mock input data
    bins = [bin1, bin2]

    output_file = tmp_path / "output.tsv"

    # Call the function
    io_manager.write_bin_info(bins, output_file, add_contigs=True)

    # Check if the file was created and its content matches the expected output
    assert Path(output_file).exists()

    with open(output_file, "r") as f:
        content = f.read()
        assert "bin_id\torigin\tname\tcompleteness\tcontamination\tscore\tsize\tN50\tcontig_count\tcontigs" in content
        assert "1\torigin1\tname1\t90\t5\t80\t1000\t500\t2\tcontig1;contig3" in content
        assert "2\torigin2\tname2\t85\t8\t75\t1200\t600\t2\tcontig2;contig4" in content




def test_write_bins_fasta(tmp_path, bin1, bin2):
    # Mock input data
    contigs_fasta = tmp_path / "contigs.fasta"
    contigs_fasta_content = (
        ">contig1\nACGT\n>contig2\nTGCA\n>contig3\nAAAA\n>contig4\nCCCC\n"
    )
    contigs_fasta.write_text(contigs_fasta_content)

    selected_bins = [bin1, bin2]

    outdir = tmp_path / "output_bins"
    outdir.mkdir()

    # Call the function
    io_manager.write_bins_fasta(selected_bins, str(contigs_fasta), str(outdir))

    # Check if the files were created and their content matches the expected output
    assert (outdir / "bin_1.fa").exists()
    assert (outdir / "bin_2.fa").exists()

    with open(outdir / "bin_1.fa", "r") as bin1_file:
        assert bin1_file.read() == ">contig1\nACGT\n>contig3\nAAAA\n"

    with open(outdir / "bin_2.fa", "r") as bin2_file:
        assert bin2_file.read() == ">contig2\nTGCA\n>contig4\nCCCC\n"
    

def test_check_contig_consistency_error():
    # Mock input data
    contigs_from_assembly = ['contig1', 'contig2', 'contig3']
    contigs_from_bins = ['contig2', 'contig3', 'contig4']
    assembly_file = "assembly.fasta"
    elsewhere_file = "external.fasta"

    with pytest.raises(AssertionError):
        # Call the function
        io_manager.check_contig_consistency(
            contigs_from_assembly, contigs_from_bins, assembly_file, elsewhere_file
        )

def test_check_contig_consistency_no_error():
    # Mock input data
    contigs_from_assembly = ['contig1', 'contig2', 'contig3', 'contig4']
    contigs_from_bins = ['contig1', 'contig2', 'contig3']
    assembly_file = "assembly.fasta"
    elsewhere_file = "external.fasta"

    io_manager.check_contig_consistency(
            contigs_from_assembly, contigs_from_bins, assembly_file, elsewhere_file
        )

@pytest.fixture
def temp_files(tmp_path):
    # Create temporary files for testing
    faa_file = tmp_path / "test_protein.faa"
    diamond_result_file = tmp_path / "test_diamond_result.txt"
    faa_file.touch()
    diamond_result_file.touch()
    yield str(faa_file), str(diamond_result_file)

def test_check_resume_file_exists(temp_files, caplog):
    # Test when both files exist
    faa_file, diamond_result_file = temp_files
    io_manager.check_resume_file(faa_file, diamond_result_file)
    assert "Protein file" not in caplog.text
    assert "Diamond result file" not in caplog.text

def test_check_resume_file_missing_faa(temp_files, caplog):
    # Test when faa_file is missing
    _, diamond_result_file = temp_files
    with pytest.raises(FileNotFoundError):
        io_manager.check_resume_file("nonexistent.faa", diamond_result_file)
    assert "Protein file" in caplog.text
    assert "Diamond result file" not in caplog.text

def test_check_resume_file_missing_diamond(temp_files, caplog):
    # Test when diamond_result_file is missing
    faa_file, _ = temp_files
    with pytest.raises(FileNotFoundError):
        io_manager.check_resume_file(faa_file, "nonexistent_diamond_result.txt")
    assert "Protein file" not in caplog.text
    assert "Diamond result file" in caplog.text



from binette import contig_manager
import pyfastx
import pytest

# Parses a valid FASTA file and returns a pyfastx.Fasta object.
def test_valid_fasta_file():
    # Arrange
    fasta_file = "tests/contigs.fasta"

    # Act
    result = contig_manager.parse_fasta_file(fasta_file)

    # Assert
    assert isinstance(result, pyfastx.Fasta)

# Parses an invalid FASTA file and raises an exception.
def test_invalid_fasta_file():
    # 
    fasta_file = "tests/contig_manager_test.py"

    # Act and Assert
    with pytest.raises(RuntimeError):
        contig_manager.parse_fasta_file(fasta_file)
        

# The function returns a tuple containing two dictionaries.
def test_returns_tuple():
    contigs = ['contig1', 'contig2', 'contig3']
    result = contig_manager.make_contig_index(contigs)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert isinstance(result[0], dict)
    assert isinstance(result[1], dict)

# The first dictionary maps contig names to their index.
def test_contig_to_index_mapping():
    contigs = ['contig1', 'contig2', 'contig3']
    result = contig_manager.make_contig_index(contigs)
    contig_to_index = result[0]
    assert isinstance(contig_to_index, dict)
    assert len(contig_to_index) == len(contigs)
    for contig in contigs:
        assert contig in contig_to_index
        assert isinstance(contig_to_index[contig], int)

# The function returns a dictionary with the same number of items as the input dictionary.
def test_same_number_of_items():
    contig_to_index = {'contig1': 0, 'contig2': 1, 'contig3': 2}
    contig_to_info = {'contig1': 'info1', 'contig2': 'info2', 'contig3': 'info3'}
    expected_result = {0: 'info1', 1: 'info2', 2: 'info3'}

    result = contig_manager.apply_contig_index(contig_to_index, contig_to_info)

    assert len(result) == len(expected_result)


import subprocess
import logging
import sys
import shutil
import re
import pandas as pd
from collections import Counter

from checkm2 import keggData


def get_checkm2_db() -> str:
    """
    Get the path to the CheckM2 database.

    :return: The path to the CheckM2 database.
    """
    if shutil.which("checkm2") is None:
        logging.error("Make sure checkm2 is on your system path.")
        sys.exit(1)

    checkm2_database_raw = subprocess.run(["checkm2", "database", "--current"], text=True, stderr=subprocess.PIPE)

    if checkm2_database_raw.returncode != 0:
        logging.error(f"Something went wrong with checkm2:\n=======\n{checkm2_database_raw.stderr}========")
        sys.exit(1)

    reg_result = re.search("INFO: (/.*.dmnd)", checkm2_database_raw.stderr)

    try:
        db_path = reg_result.group(1)
    except AttributeError:
        logging.error(f"Something went wrong when retrieving checkm2 db path:\n{checkm2_database_raw.stderr}")
        sys.exit(1)

    return db_path



def check_tool_exists(tool_name: str):
    """
    Check if a specified tool is on the system's PATH.

    :param tool_name: The name of the tool to check for.
    :type tool_name: str
    :raises FileNotFoundError: If the tool is not found on the system's PATH.
    """
    if shutil.which(tool_name) is None:
        raise FileNotFoundError(f"The '{tool_name}' tool is not found on your system PATH.")


def run(
    faa_file: str, output: str, db: str, log: str, threads: int = 1, query_cover: int = 80, subject_cover: int = 80,
    percent_id: int = 30, evalue: float = 1e-05, low_mem: bool = False):
    """
    Run Diamond with specified parameters.

    :param faa_file: Path to the input protein sequence file (FASTA format).
    :param output: Path to the Diamond output file.
    :param db: Path to the Diamond database.
    :param log: Path to the log file.
    :param threads: Number of CPU threads to use (default is 1).
    :param query_cover: Minimum query coverage percentage (default is 80).
    :param subject_cover: Minimum subject coverage percentage (default is 80).
    :param percent_id: Minimum percent identity (default is 30).
    :param evalue: Maximum e-value threshold (default is 1e-05).
    :param low_mem: Use low memory mode if True (default is False).
    """
    check_tool_exists("diamond")

    blocksize = 0.5 if low_mem else 2

    cmd = (
        "diamond blastp --outfmt 6 --max-target-seqs 1 "
        f"--query {faa_file} "
        f"-o {output} "
        f"--threads {threads} "
        f"--db {db} "
        f"--query-cover {query_cover} "
        f"--subject-cover {subject_cover} "
        f"--id {percent_id} "
        f"--evalue {evalue} --block-size {blocksize} 2> {log}"
    )

    logging.info("Running diamond")
    logging.info(cmd)

    run = subprocess.run(cmd, shell=True)

    if run.returncode != 0:
        logging.error(f"An error occurred while running DIAMOND. Check log file: {log}")
        sys.exit(1)

    logging.info("Finished Running DIAMOND")

def get_contig_to_kegg_id(diamond_result_file: str) -> dict:
    """
    Get a dictionary mapping contig IDs to KEGG annotations from a Diamond result file.

    :param diamond_result_file: Path to the Diamond result file.
    :return: A dictionary mapping contig IDs to KEGG annotations.
    """
    diamon_results_df = pd.read_csv(diamond_result_file, sep="\t", usecols=[0, 1], names=["ProteinID", "annotation"])
    diamon_results_df[["Ref100_hit", "Kegg_annotation"]] = diamon_results_df["annotation"].str.split(
        "~", n=1, expand=True
    )

    KeggCalc = keggData.KeggCalculator()
    defaultKOs = KeggCalc.return_default_values_from_category("KO_Genes")

    diamon_results_df = diamon_results_df.loc[diamon_results_df["Kegg_annotation"].isin(defaultKOs.keys())]
    diamon_results_df["contig"] = diamon_results_df["ProteinID"].str.split("_", n=-1).str[:-1].str.join("_")

    contig_to_kegg_counter = (
        diamon_results_df.groupby("contig").agg({"Kegg_annotation": Counter}).reset_index()
    )

    contig_to_kegg_counter = dict(zip(contig_to_kegg_counter["contig"], contig_to_kegg_counter["Kegg_annotation"]))

    return contig_to_kegg_counter



import re
import subprocess
import shutil
import sys
import logging

from unittest.mock import patch, MagicMock

import subprocess
import sys
import logging
import pytest

from binette import diamond

import pandas as pd
from collections import Counter

class CompletedProcess:
    def __init__(self, returncode, stderr):
        self.returncode = returncode
        self.stderr = stderr


def mock_shutil_which(*args, **kwargs):
    if args[0] == "checkm2":
        return "checkm2"

def test_get_checkm2_db_no_checkm2(monkeypatch):
    # Mocking shutil.which
    def mock_shutil_which_none(*args, **kwargs):
        return None

    monkeypatch.setattr(shutil, "which", mock_shutil_which_none)

    # Call the function
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        diamond.get_checkm2_db()

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

def test_get_checkm2_db_with_success(monkeypatch):

    def mock_subprocess_run(*args, **kwargs):

        # Simulating the behavior of checkm2 command
        if args[0][0] == "checkm2" and args[0][1] == "database" and args[0][2] == "--current":
            return CompletedProcess(0, "INFO: /mocked/path/to/checkm2.dmnd")
            
    monkeypatch.setattr(subprocess, "run", mock_subprocess_run)
    monkeypatch.setattr(shutil, "which", mock_shutil_which)

    # Call the function
    result = diamond.get_checkm2_db()

    expected_path = "/mocked/path/to/checkm2.dmnd"
    assert result == expected_path

def test_get_checkm2_db_checkm2_exit_error(monkeypatch):

    def mock_subprocess_run(*args, **kwargs):

        # Simulating the behavior of checkm2 command
        if args[0][0] == "checkm2" and args[0][1] == "database" and args[0][2] == "--current":
            return CompletedProcess(2, "")
            

    monkeypatch.setattr(subprocess, "run", mock_subprocess_run)
    monkeypatch.setattr(shutil, "which", mock_shutil_which)

    # Call the function
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        diamond.get_checkm2_db()

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1

def test_get_checkm2_db_wrong_path_format(monkeypatch):

    def mock_subprocess_run(*args, **kwargs):

        # Simulating the behavior of checkm2 command
        if args[0][0] == "checkm2" and args[0][1] == "database" and args[0][2] == "--current":
            return CompletedProcess(0, "UNEXPECTED PATH FORMAT RETURNED BY CHECKM2: /mocked/path/to/checkm2.dmnd")
            
    monkeypatch.setattr(subprocess, "run", mock_subprocess_run)
    monkeypatch.setattr(shutil, "which", mock_shutil_which)

    # Call the function
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        diamond.get_checkm2_db()

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1



def test_check_tool_exists_tool_found(monkeypatch):
    # Mocking shutil.which
    def mock_shutil_which(*args, **kwargs):
        return "path/to/tool"

    monkeypatch.setattr(shutil, "which", mock_shutil_which)

    # Call the function
    try:
        diamond.check_tool_exists("existing_tool")
    except FileNotFoundError:
        pytest.fail("check_tool_exists raised FileNotFoundError unexpectedly.")


def test_check_tool_exists_tool_not_found(monkeypatch):
    # Mocking shutil.which
    def mock_shutil_which(*args, **kwargs):
        return None

    monkeypatch.setattr(shutil, "which", mock_shutil_which)

    # Call the function and expect FileNotFoundError
    with pytest.raises(FileNotFoundError):
        diamond.check_tool_exists("non_existing_tool")
        




def test_run_diamond_tool_found(monkeypatch):

    monkeypatch.setattr(sys, "exit", lambda x: None)  # Patch sys.exit to avoid test interruption

    # Mocking subprocess.run
    def mock_subprocess_run(*args, **kwargs):
        class CompletedProcess:
            def __init__(self, returncode):
                self.returncode = returncode

        # Simulating successful run of diamond command
        if args[0] == "diamond blastp --outfmt 6 --max-target-seqs 1 --query test.faa -o output.txt --threads 1 --db db --query-cover 80 --subject-cover 80 --id 30 --evalue 1e-05 --block-size 2 2> log.txt":
            return CompletedProcess(0)

    monkeypatch.setattr(subprocess, "run", mock_subprocess_run)
    monkeypatch.setattr(logging, "error", lambda x: None)  # Avoid logging during test

    # Call the function
    diamond.run(
        "test.faa", "output.txt", "db", "log.txt", threads=1, query_cover=80, subject_cover=80, percent_id=30,
        evalue=1e-05, low_mem=False
    )


def test_run_diamond_tool_not_found(monkeypatch):
    # Mocking check_tool_exists to simulate tool not found scenario
    def mock_check_tool_exists(*args, **kwargs):
        raise FileNotFoundError

    monkeypatch.setattr(logging, "error", lambda x: None)  # Avoid logging during test

    # Call the function and expect it to raise FileNotFoundError
    with patch('sys.exit') as mock_exit:
        diamond.run(
            "test.faa", "output.txt", "db", "log.txt", threads=1, query_cover=80, subject_cover=80, percent_id=30,
            evalue=1e-05, low_mem=False
        )

    mock_exit.assert_called_once_with(1)


def test_get_contig_to_kegg_id():
    # Mock input data
    diamond_result_file = "dummy_diamond_results.txt"

    # Mocked dataframe representing the data read from the Diamond result file
    mocked_data = {
        "ProteinID": ["contig1_protein1", "contig1_protein2", "contig2_protein1", "contig2_protein2"],
        "annotation": ["protein1_annotation~K12345", "protein2_annotation~K67890", "protein3_annotation~K23456", "protein4_annotation~K66666"]
    }
    mocked_df = pd.DataFrame(mocked_data)

    # Mocked return values for keggData.KeggCalculator() and KeggCalc.return_default_values_from_category()
    class MockedKeggCalculator:
        def return_default_values_from_category(self, category):
            return {"K12345": 2, "K67890": 1, "K23456":3}

    mocked_kegg_calculator =  MockedKeggCalculator()

    # Mocking relevant functions and classes used within the function
    with patch("pandas.read_csv", return_value=mocked_df), \
         patch("checkm2.keggData.KeggCalculator", return_value=mocked_kegg_calculator):
        
        # Call the function
        result = diamond.get_contig_to_kegg_id(diamond_result_file)

    # K66666 is not in return_default_values_from_category so it won't be kept in result kegg

    # Define the expected output based on the mocked data
    expected_result = {
        "contig1": Counter({"K12345": 1, "K67890": 1}),
        "contig2": Counter({"K23456": 1})
    }

    # Check if the function output matches the expected result
    assert result == expected_result


# Additional tests can be added to cover more edge cases and scenarios.



from itertools import islice
from binette import bin_quality

from collections import Counter
import pandas as pd
from unittest.mock import Mock, patch

from unittest.mock import Mock, patch
from binette.bin_quality import (
    Bin,
    add_bin_metrics,
    assess_bins_quality_by_chunk,
    assess_bins_quality,
     chunks,
     get_diamond_feature_per_bin_df,
     get_bins_metadata_df)

from checkm2 import keggData, modelPostprocessing, modelProcessing

from unittest.mock import Mock, patch, MagicMock

def test_compute_N50():
    assert bin_quality.compute_N50([50]) == 50
    assert bin_quality.compute_N50([0]) == 0
    assert bin_quality.compute_N50([]) == 0
    assert bin_quality.compute_N50([30, 40, 30]) == 30
    assert bin_quality.compute_N50([1, 3, 3, 4, 5, 5, 6, 9, 10, 24]) == 9



def test_chunks():
    # Test case 1
    iterable_1 = [1, 2, 3, 4, 5, 6]
    size_1 = 2
    expected_output_1 = [(1, 2), (3, 4), (5, 6)]

    result_1 = list(chunks(iterable_1, size_1))
    assert result_1 == expected_output_1

    # Test case 2
    iterable_2 = [10, 20, 30, 40, 50]
    size_2 = 3
    expected_output_2 = [(10, 20, 30), (40, 50)]

    result_2 = list(chunks(iterable_2, size_2))
    assert result_2 == expected_output_2

    # Test case 3 (Empty iterable)
    iterable_3 = []
    size_3 = 5
    expected_output_3 = []

    result_3 = list(chunks(iterable_3, size_3))
    assert result_3 == expected_output_3

    # Test case 4 (Iterable length less than chunk size)
    iterable_4 = [100, 200, 300]
    size_4 = 5
    expected_output_4 = [(100, 200, 300)]

    result_4 = list(chunks(iterable_4, size_4))
    assert result_4 == expected_output_4
    


class Bin:
    def __init__(self, bin_id, contigs):
        self.id = bin_id
        self.contigs = contigs
        self.length = 0  # Mocking the add_length method
        self.N50 = 0  # Mocking the add_N50 method

    def add_length(self, length):
        self.length = length

    def add_N50(self, N50):
        self.N50 = N50

    def add_N50(self, N50):
        self.N50 = N50
    
    def add_quality(self, comp, cont, weight):

        self.completeness = comp
        self.contamination = cont
        self.score = comp - weight * cont
        
def test_get_bins_metadata_df():
    # Mock input data
    bins = [
        Bin(1, ['contig1', 'contig3']),
        Bin(2, ['contig2'])
    ]

    contig_to_cds_count = {'contig1': 10, 'contig2': 45, 'contig3': 20, 'contig4': 25}
    contig_to_aa_counter = {'contig1': Counter({'A': 5, 'D': 10}), 'contig2': Counter({'G': 8, 'V': 12, 'T': 2}), 
                            'contig3': Counter({'D': 8, 'Y': 12})}
    contig_to_aa_length = {'contig1': 1000, 'contig2': 1500, 'contig3': 2000, 'contig4': 2500}

    # Call the function
    result_df = bin_quality.get_bins_metadata_df(bins, contig_to_cds_count, contig_to_aa_counter, contig_to_aa_length)

    # Define expected values based on the provided input
    expected_columns = [
        'Name', 'A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y',
        'AALength', 'CDS'
    ]
    expected_index = [1, 2]

    expected_values = [
        [1, 5, 0, 18, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 3000, 30],
        [2, 0, 0, 0, 0, 0, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 12, 0, 0, 1500, 45]
    ]

    # Check if the generated DataFrame matches the expected DataFrame
    assert result_df.columns.tolist() == expected_columns
    assert result_df.index.tolist() == expected_index
    assert result_df.values.tolist() == expected_values




def test_get_diamond_feature_per_bin_df():
    # Mock input data
    bins = [
        Bin(1, ['contig1', 'contig2']),
        Bin(2, ['contig3', 'contig4'])
    ]

    contig_to_kegg_counter = {
        'contig1': Counter({'K01810': 5, 'K15916': 7}),
        'contig2': Counter({'K01810': 10}),
        'contig3': Counter({'K00918': 8}),
    }

    # Call the function
    result_df, default_ko_count = bin_quality.get_diamond_feature_per_bin_df(bins, contig_to_kegg_counter)

    expected_index = [1, 2]
    
    
    assert result_df.index.tolist() == expected_index
    assert result_df.loc[1,"K01810"] == 15 # in bin1 from contig 1 and 2
    assert result_df.loc[1,"K15916"] == 7 # in bin1 from contig 1 
    assert result_df.loc[2,"K01810"] == 0 # this ko is not in any contig of bin 2
    assert result_df.loc[2,"K00918"] == 8 # in bin2 from contig 3



def test_add_bin_size_and_N50():
    # Mock input data
    bins = [
        Bin(1, ['contig1', 'contig2']),
        Bin(2, ['contig3'])
    ]

    contig_to_size = {
        'contig1': 1000,
        'contig2': 1500,
        'contig3': 2000,
    }

    # Call the function
    bin_quality.add_bin_size_and_N50(bins, contig_to_size)

    # Assertions to verify if add_length and add_N50 were called with the correct values
    assert bins[0].length == 2500
    assert bins[0].N50 == 1500 
    assert bins[1].length == 2000
    assert bins[1].N50 == 2000 




def mock_modelProcessor(thread):
    return "mock_modelProcessor"

def test_add_bin_metrics(monkeypatch):
    # Mock input data
    bins = [
        Bin(1, ['contig1', 'contig2']),
        Bin(2, ['contig3'])
    ]

    contig_info = {
        # Add mocked contig information here as needed
        "contig_to_kegg_counter":{},
        "contig_to_cds_count":{},
        "contig_to_aa_counter":{},
        "contig_to_aa_length":{},
        "contig_to_length":{},
    }

    contamination_weight = 0.5
    threads = 1
    

    monkeypatch.setattr(modelPostprocessing, "modelProcessor", mock_modelProcessor)


    # Mock the functions called within add_bin_metrics
    with patch('binette.bin_quality.add_bin_size_and_N50') as mock_add_bin_size_and_N50, \
         patch('binette.bin_quality.assess_bins_quality_by_chunk') as mock_assess_bins_quality_by_chunk:

        add_bin_metrics(bins, contig_info, contamination_weight, threads)

        # Assertions to check if functions were called with the expected arguments
        mock_add_bin_size_and_N50.assert_called_once_with(bins, contig_info["contig_to_length"])
        mock_assess_bins_quality_by_chunk.assert_called_once_with(
            bins,
            contig_info["contig_to_kegg_counter"],
            contig_info["contig_to_cds_count"],
            contig_info["contig_to_aa_counter"],
            contig_info["contig_to_aa_length"],
            contamination_weight,
            "mock_modelProcessor",  # Mocked postProcessor object
        )

def test_assess_bins_quality_by_chunk(monkeypatch):
    # Prepare input data for testing
    bins = [
        Bin(1, ['contig1', 'contig2']),
        Bin(2, ['contig3', 'contig4']),
        Bin(3, ['contig3', 'contig4'])
    ]

    contig_to_kegg_counter = {}
    contig_to_cds_count = {}
    contig_to_aa_counter = {}
    contig_to_aa_length = {}  
    contamination_weight = 0.5 

    # Mocking postProcessor object

    monkeypatch.setattr(modelPostprocessing, "modelProcessor", mock_modelProcessor)


    # Mock the functions called within add_bin_metrics
    with patch('binette.bin_quality.assess_bins_quality') as mock_assess_bins_quality:

        assess_bins_quality_by_chunk(
                                    bins,
                                    contig_to_kegg_counter,
                                    contig_to_cds_count,
                                    contig_to_aa_counter,
                                    contig_to_aa_length,
                                    contamination_weight,
                                    postProcessor=None,
                                    threads=1,
                                    chunk_size = 3 
                                )

        # Chunk size > number of bin so only one chunk
        mock_assess_bins_quality.assert_called_once_with(
            bins= set(bins),
            contig_to_kegg_counter= contig_to_kegg_counter,
            contig_to_cds_count=contig_to_cds_count,
            contig_to_aa_counter=contig_to_aa_counter,
            contig_to_aa_length=contig_to_aa_length,
            contamination_weight=contamination_weight,
            postProcessor=None,
            threads=1
        )

    # Mock the functions called within add_bin_metrics
    with patch('binette.bin_quality.assess_bins_quality') as mock_assess_bins_quality:

        assess_bins_quality_by_chunk(
                                    bins,
                                    contig_to_kegg_counter,
                                    contig_to_cds_count,
                                    contig_to_aa_counter,
                                    contig_to_aa_length,
                                    contamination_weight,
                                    postProcessor=None,
                                    threads=1,
                                    chunk_size = 2
                                )

        # Chunk size < number of bin so  2 chunks with [bin1,bin2] and [bin3]
        assert mock_assess_bins_quality.call_count == 2



from unittest.mock import patch, MagicMock
import numpy as np
import pandas as pd
from checkm2 import keggData, modelPostprocessing, modelProcessing


def test_assess_bins_quality():
    # Prepare mock input data for testing
    bins = [
        Bin(1, ['contig1', 'contig2']),
        Bin(2, ['contig3', 'contig4'])
    ]

    contig_to_kegg_counter = {}
    contig_to_cds_count = {}
    contig_to_aa_length = {}
    contig_to_aa_counter = {}
    contamination_weight = 0.5


    # Call the function being tested
    assess_bins_quality(
        bins,
        contig_to_kegg_counter,
        contig_to_cds_count,
        contig_to_aa_counter,
        contig_to_aa_length,
        contamination_weight
    )



    # Verify the expected calls to add_quality for each bin object
    for bin_obj in bins:
        assert bin_obj.completeness is not None
        assert bin_obj.contamination is not None
        assert bin_obj.score is not None
        assert bin_obj.score == bin_obj.completeness - bin_obj.contamination * contamination_weight



import pytest
import logging
from binette.main import log_selected_bin_info, select_bins_and_write_them, manage_protein_alignement, parse_input_files, parse_arguments, init_logging, main
from binette.bin_manager import Bin
from binette import diamond
import os
import sys
from unittest.mock import patch

from collections import Counter
from tests.bin_manager_test import create_temp_bin_directories, create_temp_bin_files

@pytest.fixture
def bins():
    b1 = Bin(contigs={"contig1"}, origin="set1", name="bin1")
    b2 = Bin(contigs={"contig3"}, origin="set1", name="bin2")
    b3 = Bin(contigs={"contig3", "contig2"}, origin="set1", name="bin3")

    b1.add_quality(100, 0, 0)
    b2.add_quality(95, 10, 0)
    b3.add_quality(70, 20, 0)

    return [b1, b2, b3]

def test_log_selected_bin_info(caplog, bins):

    caplog.set_level(logging.INFO)


    hq_min_completeness = 85
    hq_max_conta = 15

    # Call the function
    log_selected_bin_info(bins, hq_min_completeness, hq_max_conta)

    # Check if the logs contain expected messages
    expected_logs ="2/3 selected bins have a high quality (completeness >= 85 and contamination <= 15)."


    assert expected_logs in caplog.text


def test_select_bins_and_write_them(tmp_path, tmpdir, bins):
    # Create temporary directories and files for testing
    outdir = tmpdir.mkdir("test_outdir")
    contigs_fasta = os.path.join(str(outdir), "contigs.fasta")
    final_bin_report = os.path.join(str(outdir), "final_bin_report.tsv")

    index_to_contig={"contig1":"contig1", "contig2": "contig2", "contig3":"contig3"}

    contigs_fasta = tmp_path / "contigs.fasta"
    contigs_fasta_content = (
    ">contig1\nACGT\n>contig2\nTGCA\n>contig3\nAAAA\n>contig4\nCCCC\n"
    )
    contigs_fasta.write_text(contigs_fasta_content)

    b1, b2, b3 = bins


    # Run the function with test data
    selected_bins = select_bins_and_write_them(
        set(bins), str(contigs_fasta), final_bin_report, min_completeness=60, index_to_contig=index_to_contig, outdir=str(outdir), debug=False
    )

    # Assertions to check the function output or file existence
    assert isinstance(selected_bins, list)
    assert os.path.isfile(final_bin_report)
    assert selected_bins == bins[:2] # The third bin is overlapping with the second one and has a worse score so it is not selected.

    with open(outdir / f"final_bins/bin_{b1.id}.fa", "r") as bin1_file:
        assert bin1_file.read() == ">contig1\nACGT\n"

    with open(outdir / f"final_bins/bin_{b2.id}.fa", "r") as bin2_file:
        assert bin2_file.read() == ">contig3\nAAAA\n"

    assert not os.path.isfile(outdir / f"final_bins/bin_{b3.id}.fa")



def test_manage_protein_alignement_resume(tmp_path):
    # Create temporary directories and files for testing

    faa_file = tmp_path / "proteins.faa"
    faa_file_content = (
    ">contig1_1\nACGT\n>contig2_1\nTGCA\n>contig2_2\nAAAA\n>contig3_1\nCCCC\n"
    )

    contig_to_length={"contig1":40, "contig2":80, "contig3":20}

    faa_file.write_text(faa_file_content)

    contig_to_kegg_id = {
        "contig1": Counter({"K12345": 1, "K67890": 1}),
        "contig2": Counter({"K23456": 1})
    }


    with patch("binette.diamond.get_contig_to_kegg_id", return_value=contig_to_kegg_id):
        
        # Call the function

        # Run the function with test data
        contig_to_kegg_counter, contig_to_genes = manage_protein_alignement(
            faa_file=str(faa_file),
            contigs_fasta="contigs_fasta",
            contig_to_length=contig_to_length,
            contigs_in_bins={},
            diamond_result_file="diamond_result_file",
            checkm2_db=None,
            threads=1,
            resume=True,
            low_mem=False
        )

    # Assertions to check the function output or file existence
    assert isinstance(contig_to_genes, dict)
    assert isinstance(contig_to_kegg_counter, dict)
    assert len(contig_to_genes) == 3


def test_manage_protein_alignement_not_resume(tmpdir, tmp_path):
    # Create temporary directories and files for testing

    faa_file = tmp_path / "proteins.faa"
    faa_file_content = (
    ">contig1_1\nACGT\n>contig2_1\nTGCA\n>contig2_2\nAAAA\n>contig3_1\nCCCC\n"
    )

    contig_to_length={"contig1":40, "contig2":80, "contig3":20}

    faa_file.write_text(faa_file_content)


    contigs_fasta = os.path.join(str(tmpdir), "contigs.fasta")
    diamond_result_file = os.path.join(str(tmpdir), "diamond_results.tsv")

    contig_to_kegg_id = {
        "contig1": Counter({"K12345": 1, "K67890": 1}),
        "contig2": Counter({"K23456": 1})
    }


    with patch("binette.diamond.get_contig_to_kegg_id", return_value=contig_to_kegg_id), \
                             patch("binette.diamond.run", return_value=None):
        
        # Call the function

        contig_to_kegg_counter, contig_to_genes = manage_protein_alignement(
            faa_file=str(faa_file),
            contigs_fasta=contigs_fasta,
            contig_to_length=contig_to_length,
            contigs_in_bins={},
            diamond_result_file=diamond_result_file,
            checkm2_db=None,
            threads=1,
            resume=True,
            low_mem=False
        )

    # Assertions to check the function output or file existence
    assert isinstance(contig_to_genes, dict)
    assert isinstance(contig_to_kegg_counter, dict)
    assert len(contig_to_genes) == 3


def test_parse_input_files_bin_dirs(create_temp_bin_directories, tmp_path):

    set_name_to_bin_dir = create_temp_bin_directories
    bin_dirs = list(create_temp_bin_directories.values())

    contig2bin_tables = []

    # Create temporary directories and files for testing

    fasta_file = tmp_path / "assembly.fasta"
    fasta_file_content = (
    ">contig1\nACGT\n>contig2\nTGCA\n>contig3\nAAAA\n>contig4\nCCCC\n>contig5\nCGTCGCT\n"
    )
    fasta_file.write_text(fasta_file_content)

    # Call the function and capture the return values
    bin_set_name_to_bins, original_bins, contigs_in_bins, contig_to_length = parse_input_files(bin_dirs, contig2bin_tables, str(fasta_file))


    # # Perform assertions on the returned values
    assert isinstance(bin_set_name_to_bins, dict)
    assert isinstance(original_bins, set)
    assert isinstance(contigs_in_bins, set)
    assert isinstance(contig_to_length, dict)


    assert set(bin_set_name_to_bins) == {'1', "2"}
    assert len(original_bins) == 3
    assert contigs_in_bins == {"contig1","contig2", "contig3","contig4","contig5",}
    assert len(contig_to_length) == 5


def test_parse_input_files_bin_dirs(create_temp_bin_directories, tmp_path):

    set_name_to_bin_dir = create_temp_bin_directories
    bin_dirs = list(create_temp_bin_directories.values())

    contig2bin_tables = []

    # Create temporary directories and files for testing

    fasta_file = tmp_path / "assembly.fasta"
    fasta_file_content = (
    ">contig1\nACGT\n>contig2\nTGCA\n>contig3\nAAAA\n>contig4\nCCCC\n>contig5\nCGTCGCT\n"
    )
    fasta_file.write_text(fasta_file_content)

    # Call the function and capture the return values
    bin_set_name_to_bins, original_bins, contigs_in_bins, contig_to_length = parse_input_files(bin_dirs, contig2bin_tables, str(fasta_file))

    # # Perform assertions on the returned values
    assert isinstance(bin_set_name_to_bins, dict)
    assert isinstance(original_bins, set)
    assert isinstance(contigs_in_bins, set)
    assert isinstance(contig_to_length, dict)


    assert set(bin_set_name_to_bins) == {'1', "2"}
    assert len(original_bins) == 3
    assert contigs_in_bins == {"contig1","contig2", "contig3","contig4","contig5",}
    assert len(contig_to_length) == 5



def test_parse_arguments_required_arguments():
    # Test when only required arguments are provided
    args = parse_arguments(["-d", "folder1", "folder2", "-c", "contigs.fasta"])
    assert args.bin_dirs == ["folder1", "folder2"]
    assert args.contigs == "contigs.fasta"

def test_parse_arguments_optional_arguments():
    # Test when required and optional arguments are provided
    args = parse_arguments(["-d", "folder1", "folder2", "-c", "contigs.fasta", "--threads", "4", "--outdir", "output"])
    assert args.bin_dirs == ["folder1", "folder2"]
    assert args.contigs == "contigs.fasta"
    assert args.threads == 4
    assert args.outdir == "output"

def test_parse_arguments_invalid_arguments():
    # Test when invalid arguments are provided
    with pytest.raises(SystemExit):
        # In this case, required arguments are missing
        parse_arguments(["-t", "4"])

def test_parse_arguments_help():
    # Test the help message
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        parse_arguments(["-h"])
    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0


def test_init_logging_command_line(caplog):
    
    caplog.set_level(logging.INFO)

    init_logging(verbose=True, debug=False)
    expected_log_message = f'command line: {" ".join(sys.argv)}'
    # Check if the log message is present in the log records

    assert expected_log_message in caplog.text



def test_main_functionality_resume_when_not_possible(monkeypatch):
    # Define or mock the necessary inputs/arguments

    # Mock sys.argv to use test_args
    test_args = [
        "-d", "bin_dir1", "bin_dir2",
        "-c", "contigs.fasta",
        # ... more arguments as required ...
        "--debug",
        "--resume"
    ]
    monkeypatch.setattr(sys, 'argv', ['your_script.py'] + test_args)

    # You may also need to mock certain functions to avoid actual file operations or to simulate their behavior
    # For example, mock the functions parse_input_files, manage_protein_alignement, select_bins_and_write_them, etc.

    # Call the main function
    with pytest.raises(FileNotFoundError) as e_info:
        main()


# def test_main_functionality_(monkeypatch, create_temp_bin_directories, tmp_path):
#     # Define or mock the necessary inputs/arguments

#     bin_dirs = list(create_temp_bin_directories.values())
#     fasta_file = tmp_path / "assembly.fasta"
#     fasta_file_content = (
#     ">contig1\nACGT\n>contig2\nTGCA\n>contig3\nAAAA\n>contig4\nCCCC\n>contig5\nCGTCGCT\n"
#     )
#     fasta_file.write_text(fasta_file_content)

#     # Mock sys.argv to use test_args
#     test_args = [
#         "-d"] + bin_dirs + ["-c", str(fasta_file),
#         # ... more arguments as required ...
#         "--debug"
#     ]

#     monkeypatch.setattr(sys, 'argv', ['your_script.py'] + test_args)

#     # You may also need to mock certain functions to avoid actual file operations or to simulate their behavior
#     # For example, mock the functions parse_input_files, manage_protein_alignement, select_bins_and_write_them, etc.

#     # Call the main function
#     with patch("binette.diamond.get_contig_to_kegg_id", return_value=contig_to_kegg_id), \
#                             patch("binette.diamond.run", return_value=None):
#         result = main()
#         assert result == 0
        



import concurrent.futures as cf
import multiprocessing.pool
import logging
from collections import Counter, defaultdict
from typing import Dict, List, Iterator, Tuple

import pyfastx
import pyrodigal
from tqdm import tqdm


def get_contig_from_cds_name(cds_name: str) -> str:
    """
    Extract the contig name from a CDS name.

    :param cds_name: The name of the CDS.
    :type cds_name: str
    :return: The name of the contig.
    :rtype: str
    """
    return "_".join(cds_name.split("_")[:-1])

def predict(contigs_iterator: Iterator, outfaa: str, threads: int =1) -> Dict[str, List[str]]:
    """
    Predict open reading frames with Pyrodigal.

    :param contigs_iterator: An iterator of contig sequences.
    :param outfaa: The output file path for predicted protein sequences (in FASTA format).
    :param threads: Number of CPU threads to use (default is 1).

    :return: A dictionary mapping contig names to predicted genes.
    """
    try:
        # for version >=3 of pyrodigal
        orf_finder = pyrodigal.GeneFinder(meta="meta")
    except AttributeError:
        orf_finder = pyrodigal.OrfFinder(meta="meta")

    logging.info(f"Predicting cds sequences with Pyrodigal using {threads} threads.")
    
    with multiprocessing.pool.ThreadPool(processes=threads) as pool: 
        contig_and_genes = pool.starmap(predict_genes, ((orf_finder.find_genes, seq) for seq in contigs_iterator))

    write_faa(outfaa, contig_and_genes)
    

    contig_to_genes = {
        contig_id: [gene.translate() for gene in pyrodigal_genes]
        for contig_id, pyrodigal_genes in contig_and_genes
    }

    return contig_to_genes

def predict_genes(find_genes, seq):


    return (seq.name, find_genes(seq.seq) )


def write_faa(outfaa: str, contig_to_genes: Dict[str, List[str]]) -> None:
    """
    Write predicted protein sequences to a FASTA file.

    :param outfaa: The output file path for predicted protein sequences (in FASTA format).
    :param contig_to_genes: A dictionary mapping contig names to predicted genes.

    """
    logging.info("Writing predicted protein sequences.")
    with open(outfaa, "w") as fl:
        for contig_id, genes in contig_to_genes:
            genes.write_translations(fl, contig_id)

def parse_faa_file(faa_file: str) -> Dict[str, List]:
    """
    Parse a FASTA file containing protein sequences and organize them by contig.

    :param faa_file: Path to the input FASTA file.
    :return: A dictionary mapping contig names to lists of protein sequences.
    """
    contig_to_genes = defaultdict(list)
    for name, seq in pyfastx.Fastx(faa_file):
        contig = get_contig_from_cds_name(name)
        contig_to_genes[contig].append(seq)

    return dict(contig_to_genes)

def get_aa_composition(genes: List[str]) -> Counter:
    """
    Calculate the amino acid composition of a list of protein sequences.

    :param genes: A list of protein sequences.
    :return: A Counter object representing the amino acid composition.
    """
    aa_counter = Counter()
    for gene in genes:
        aa_counter += Counter(gene)

    return aa_counter

def get_contig_cds_metadata_flat(contig_to_genes: Dict[str, List[str]]) -> Tuple[Dict[str, int], Dict[str, Counter], Dict[str, int]]:
    """
    Calculate metadata for contigs, including CDS count, amino acid composition, and total amino acid length.

    :param contig_to_genes: A dictionary mapping contig names to lists of protein sequences.
    :return: A tuple containing dictionaries for CDS count, amino acid composition, and total amino acid length.
    """
    contig_to_cds_count = {contig: len(genes) for contig, genes in contig_to_genes.items()}

    contig_to_aa_counter = {contig: get_aa_composition(genes) for contig, genes in tqdm(contig_to_genes.items(), unit="contig")}
    logging.info("Calculating amino acid composition.")

    contig_to_aa_length = {contig: sum(counter.values()) for contig, counter in tqdm(contig_to_aa_counter.items(), unit="contig")}
    logging.info("Calculating total amino acid length.")

    return contig_to_cds_count, contig_to_aa_counter, contig_to_aa_length

def get_contig_cds_metadata(contig_to_genes: Dict[str, List[str]], threads: int) -> Tuple[Dict[str, int], Dict[str, Counter], Dict[str, int]]:
    """
    Calculate metadata for contigs in parallel, including CDS count, amino acid composition, and total amino acid length.

    :param contig_to_genes: A dictionary mapping contig names to lists of protein sequences.
    :param threads: Number of CPU threads to use.
    :return: A tuple containing dictionaries for CDS count, amino acid composition, and total amino acid length.
    """
    contig_to_cds_count = {contig: len(genes) for contig, genes in contig_to_genes.items()}

    contig_to_future = {}
    logging.info(f"Collecting contig amino acid composition using {threads} threads.")
    with cf.ProcessPoolExecutor(max_workers=threads) as tpe:
        for contig, genes in tqdm(contig_to_genes.items()):
            contig_to_future[contig] = tpe.submit(get_aa_composition, genes)

    contig_to_aa_counter = {contig: future.result() for contig, future in tqdm(contig_to_future.items(), unit="contig")}
    logging.info("Calculating amino acid composition in parallel.")

    contig_to_aa_length = {contig: sum(counter.values()) for contig, counter in  tqdm(contig_to_aa_counter.items(), unit="contig")}
    logging.info("Calculating total amino acid length in parallel.")

    contig_info = {
        "contig_to_cds_count": contig_to_cds_count,
        "contig_to_aa_counter": contig_to_aa_counter,
        "contig_to_aa_length": contig_to_aa_length,
    }

    return contig_info


import logging
import os
from collections import defaultdict
import pyfastx

import itertools
import networkx as nx
from typing import List, Dict, Iterable, Tuple, Set

class Bin:
    counter = 0

    def __init__(self, contigs: Iterable[str], origin: str, name: str) -> None:
        """
        Initialize a Bin object.

        :param contigs: Iterable of contig names belonging to the bin.
        :param origin: Origin/source of the bin.
        :param name: Name of the bin.
        """
        Bin.counter += 1

        self.origin = origin
        self.name = name
        self.id = Bin.counter
        self.contigs = set(contigs)
        self.hash = hash(str(sorted(self.contigs)))

        self.length = None
        self.N50 = None

        self.completeness = None
        self.contamination = None
        self.score = None

    def __eq__(self, other: 'Bin') -> bool:
        """
        Compare the Bin object with another object for equality.

        :param other: The object to compare with.
        :return: True if the objects are equal, False otherwise.
        """
        return self.contigs == other.contigs

    def __hash__(self) -> int:
        """
        Compute the hash value of the Bin object.

        :return: The hash value.
        """
        return self.hash

    def __str__(self) -> str:
        """
        Return a string representation of the Bin object.

        :return: The string representation of the Bin object.
        """
        return f"{self.origin}_{self.id}  ({len(self.contigs)} contigs)"

    def overlaps_with(self, other: 'Bin') -> Set[str]:
        """
        Find the contigs that overlap between this bin and another bin.

        :param other: The other Bin object.
        :return: A set of contig names that overlap between the bins.
        """
        return self.contigs & other.contigs

    def __and__(self, other: 'Bin') -> 'Bin':
        """
        Perform a logical AND operation between this bin and another bin.

        :param other: The other Bin object.
        :return: A new Bin object representing the intersection of the bins.
        """
        contigs = self.contigs & other.contigs
        name = f"{self.name} & {other.name}"
        origin = f"{self.origin} & {other.origin}"

        return Bin(contigs, origin, name)


    def add_length(self, length: int) -> None:
        """
        Add the length attribute to the Bin object if the provided length is a positive integer.

        :param length: The length value to add.
        :return: None
        """
        if isinstance(length, int) and length > 0:
            self.length = length
        else:
            raise ValueError("Length should be a positive integer.")

    def add_N50(self, n50: int) -> None:
        """
        Add the N50 attribute to the Bin object.

        :param n50: The N50 value to add.
        :return: None
        """
        if isinstance(n50, int) and n50 >= 0:
            self.N50 = n50
        else:
            raise ValueError("N50 should be a positive integer.")
        

    def add_quality(self, completeness: float, contamination: float, contamination_weight: float) -> None:
        """
        Set the quality attributes of the bin.

        :param completeness: The completeness value.
        :param contamination: The contamination value.
        :param contamination_weight: The weight assigned to contamination in the score calculation.
        """
        self.completeness = completeness
        self.contamination = contamination
        self.score = completeness - contamination_weight * contamination

    def intersection(self, *others: 'Bin') -> 'Bin':
        """
        Compute the intersection of the bin with other bins.

        :param others: Other bins to compute the intersection with.
        :return: A new Bin representing the intersection of the bins.
        """
        other_contigs = (o.contigs for o in others)
        contigs = self.contigs.intersection(*other_contigs)
        name = f"{self.name} & {' & '.join([other.name for other in others])}"
        origin = "intersec"

        return Bin(contigs, origin, name)

    def difference(self, *others: 'Bin') -> 'Bin':
        """
        Compute the difference between the bin and other bins.

        :param others: Other bins to compute the difference with.
        :return: A new Bin representing the difference between the bins.
        """
        other_contigs = (o.contigs for o in others)
        contigs = self.contigs.difference(*other_contigs)
        name = f"{self.name} - {' - '.join([other.name for other in others])}"
        origin = "diff"

        return Bin(contigs, origin, name)

    def union(self, *others: 'Bin') -> 'Bin':
        """
        Compute the union of the bin with other bins.

        :param others: Other bins to compute the union with.
        :return: A new Bin representing the union of the bins.
        """
        other_contigs = (o.contigs for o in others)
        contigs = self.contigs.union(*other_contigs)
        name = f"{self.name} | {' | '.join([other.name for other in others])}"
        origin = "union"

        return Bin(contigs, origin, name)
    

def get_bins_from_directory(bin_dir: str, set_name: str) -> List[Bin]:
    """
    Retrieves a list of Bin objects from a directory containing bin FASTA files.

    :param bin_dir: The directory path containing bin FASTA files.
    :param set_name: The name of the set the bins belong to.

    :return: A list of Bin objects created from the bin FASTA files.
    """
    bins = []

    for bin_fasta_file in os.listdir(bin_dir):

        bin_fasta_path = os.path.join(bin_dir, bin_fasta_file)
        bin_name = bin_fasta_file

        contigs = {name for name, _ in pyfastx.Fasta(bin_fasta_path, build_index=False)}

        bin_obj = Bin(contigs, set_name, bin_name)

        bins.append(bin_obj)

    return bins



def parse_bin_directories(bin_name_to_bin_dir: Dict[str, str]) -> Dict[str, list]:
    """
    Parses multiple bin directories and returns a dictionary mapping bin names to a list of Bin objects.

    :param bin_name_to_bin_dir: A dictionary mapping bin names to their respective bin directories.

    :return: A dictionary mapping bin names to a list of Bin objects created from the bin directories.
    """
    bin_name_to_bins = {}

    for name, bin_dir in bin_name_to_bin_dir.items():
        bin_name_to_bins[name] = get_bins_from_directory(bin_dir, name)

    return bin_name_to_bins


def parse_contig2bin_tables(bin_name_to_bin_tables: Dict[str, str]) -> Dict[str, list]:
    """
    Parses multiple contig-to-bin tables and returns a dictionary mapping bin names to a list of Bin objects.

    :param bin_name_to_bin_tables: A dictionary mapping bin names to their respective contig-to-bin tables.

    :return: A dictionary mapping bin names to a list of Bin objects created from the contig-to-bin tables.
    """
    bin_name_to_bins = {}

    for name, contig2bin_table in bin_name_to_bin_tables.items():
        bin_name_to_bins[name] = get_bins_from_contig2bin_table(contig2bin_table, name)

    return bin_name_to_bins


def get_bins_from_contig2bin_table(contig2bin_table: str, set_name: str) -> List[Bin]:
    """
    Retrieves a list of Bin objects from a contig-to-bin table.

    :param contig2bin_table: The path to the contig-to-bin table.
    :param set_name: The name of the set the bins belong to.

    :return: A list of Bin objects created from the contig-to-bin table.
    """
    bin_name2contigs = defaultdict(set)
    with open(contig2bin_table) as fl:
        for line in fl:
            if line.startswith("#") or line.startswith("@"):
                logging.debug(f"Ignoring a line from {contig2bin_table}: {line}")
                continue
            contig_name = line.strip().split("\t")[0]
            bin_name = line.strip().split("\t")[1]
            bin_name2contigs[bin_name].add(contig_name)

    bins = []
    for bin_name, contigs in bin_name2contigs.items():
        bin_obj = Bin(contigs, set_name, bin_name)
        bins.append(bin_obj)
    return bins


def from_bin_sets_to_bin_graph(bin_name_to_bin_set: Dict[str, set]) -> nx.Graph:
    """
    Creates a bin graph from a dictionary of bin sets.

    :param bin_name_to_bin_set: A dictionary mapping bin names to their respective bin sets.

    :return: A networkx Graph representing the bin graph of overlapping bins.
    """
    G = nx.Graph()

    for set1_name, set2_name in itertools.combinations(bin_name_to_bin_set, 2):
        set1 = bin_name_to_bin_set[set1_name]
        set2 = bin_name_to_bin_set[set2_name]

        for bin1, bin2 in itertools.product(set1, set2):

            if bin1.overlaps_with(bin2):
                G.add_edge(bin1, bin2)
    return G



def get_all_possible_combinations(clique: Iterable) -> Iterable[Tuple]:
    """
    Generates all possible combinations of elements from a given clique.

    :param clique: An iterable representing a clique.

    :return: An iterable of tuples representing all possible combinations of elements from the clique.
    """
    return (c for r in range(2, len(clique) + 1) for c in itertools.combinations(clique, r))


def get_intersection_bins(G: nx.Graph) -> Set[Bin]:
    """
    Retrieves the intersection bins from a given graph.

    :param G: A networkx Graph representing the graph.

    :return: A set of Bin objects representing the intersection bins.
    """
    intersect_bins = set()

    for clique in nx.clique.find_cliques(G):
        bins_combinations = get_all_possible_combinations(clique)
        for bins in bins_combinations:
            if max((b.completeness for b in bins)) < 20:
                logging.debug("completeness is not good enough to create a new bin on intersection")
                logging.debug(f"{[(str(b), b.completeness, b.contamination)  for b in bins]}")
                continue

            intersec_bin = bins[0].intersection(*bins[1:])

            if intersec_bin.contigs:
                intersect_bins.add(intersec_bin)

    return intersect_bins


def get_difference_bins(G: nx.Graph) -> Set[Bin]:
    """
    Retrieves the difference bins from a given graph.

    :param G: A networkx Graph representing the graph.

    :return: A set of Bin objects representing the difference bins.
    """
    difference_bins = set()
    
    for clique in nx.clique.find_cliques(G):

        bins_combinations = get_all_possible_combinations(clique)
        for bins in bins_combinations:

            for bin_a in bins:

                bin_diff = bin_a.difference(*(b for b in bins if b != bin_a))
                if bin_a.completeness < 20:
                    logging.debug(f"completeness of {bin_a} is not good enough to do difference... ")
                    logging.debug(f"{[(str(b), b.completeness, b.contamination)  for b in bins]}")
                    continue

                if bin_diff.contigs:
                    difference_bins.add(bin_diff)

    return difference_bins


def get_union_bins(G: nx.Graph, max_conta: int = 50) -> Set[Bin]:
    """
    Retrieves the union bins from a given graph.

    :param G: A networkx Graph representing the graph of bins.
    :param max_conta: Maximum allowed contamination value for a bin to be included in the union.

    :return: A set of Bin objects representing the union bins.
    """
    union_bins = set()
    for clique in nx.clique.find_cliques(G):
        bins_combinations = get_all_possible_combinations(clique)
        for bins in bins_combinations:
            if max((b.contamination for b in bins)) > max_conta:
                logging.debug("Some bin are too contaminated to make a useful union bin")
                logging.debug(f"{[(str(b), b.completeness, b.contamination)  for b in bins]}")
                continue

            bins = set(bins)
            bin_a = bins.pop()
            bin_union = bin_a.union(*bins)

            if bin_union.contigs:
                union_bins.add(bin_union)

    return union_bins


def select_best_bins(bins: List[Bin]) -> List[Bin]:
    """
    Selects the best bins from a list of bins based on their scores, N50 values, and IDs.

    :param bins: A list of Bin objects.

    :return: A list of selected Bin objects.
    """

    logging.info("Sorting bins")
    # Sort on score, N50, and ID. Smaller ID values are preferred to select original bins first.
    sorted_bins = sorted(bins, key=lambda x: (x.score, x.N50, -x.id), reverse=True)

    logging.info("Selecting bins")
    selected_bins = []
    for b in sorted_bins:
        if b in bins:
            overlapping_bins = {b2 for b2 in bins if b.overlaps_with(b2)}
            bins -= overlapping_bins

            selected_bins.append(b)

    logging.info(f"Selected {len(selected_bins)} bins")
    return selected_bins


def dereplicate_bin_sets(bin_sets):
    """
    Dereplicates bins from different bin sets to obtain a non-redundant bin set.

    :param bin_sets: A list of bin sets.

    :return: A set of non-redundant bins.
    """
    return set().union(*bin_sets)


def get_contigs_in_bins(bins: List[Bin]) -> Set[str]:
    """
    Retrieves all contigs present in the given list of bins.

    :param bins: A list of Bin objects.

    :return: A set of contigs present in the bins.
    """
    return set().union(*(b.contigs for b in bins))


def rename_bin_contigs(bins: List[Bin], contig_to_index: dict):
    """
    Renames the contigs in the bins based on the provided mapping.

    :param bins: A list of Bin objects.
    :param contig_to_index: A dictionary mapping old contig names to new index names.
    """
    for b in bins:
        b.contigs = {contig_to_index[contig] for contig in b.contigs}
        b.hash = hash(str(sorted(b.contigs)))

def create_intermediate_bins(bin_set_name_to_bins: Dict[str, Set[Bin]]) -> Set[Bin]:
    """
    Creates intermediate bins from a dictionary of bin sets.

    :param bin_set_name_to_bins: A dictionary mapping bin set names to corresponding bins.

    :return: A set of intermediate bins created from intersections, differences, and unions.
    """
    logging.info("Making bin graph...")
    connected_bins_graph = from_bin_sets_to_bin_graph(bin_set_name_to_bins)

    logging.info("Creating intersection bins...")
    intersection_bins = get_intersection_bins(connected_bins_graph)
    logging.info(f"{len(intersection_bins)} bins created on intersections.")

    logging.info("Creating difference bins...")
    difference_bins = get_difference_bins(connected_bins_graph)
    logging.info(f"{len(difference_bins)} bins created based on symmetric difference.")

    logging.info("Creating union bins...")
    union_bins = get_union_bins(connected_bins_graph)
    logging.info(f"{len(union_bins)} bins created on unions.")

    return difference_bins | intersection_bins | union_bins




#!/usr/bin/env python3
import concurrent.futures as cf
import logging
import os
from collections import Counter
from itertools import islice
from typing import Dict, Iterable, List, Tuple, Iterator

import numpy as np
import pandas as pd

# Suppress unnecessary TensorFlow warnings
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"
logging.getLogger("tensorflow").setLevel(logging.FATAL)


from checkm2 import keggData, modelPostprocessing, modelProcessing
from binette.bin_manager import Bin

def get_bins_metadata_df(bins: List, contig_to_cds_count: Dict[str, int], contig_to_aa_counter: Dict[str, Counter], contig_to_aa_length: Dict[str, int]) -> pd.DataFrame:
    """
    Generate a DataFrame containing metadata for a list of bins.

    :param bins: A list of bin objects.
    :param contig_to_cds_count: A dictionary mapping contig names to CDS counts.
    :param contig_to_aa_counter: A dictionary mapping contig names to amino acid composition (Counter object).
    :param contig_to_aa_length: A dictionary mapping contig names to total amino acid length.
    :return: A DataFrame containing bin metadata.
    """

    metadata_order = keggData.KeggCalculator().return_proper_order("Metadata")

    # Get bin metadata
    bin_metadata_list = []
    for bin_obj in bins:
        bin_metadata = {
            "Name": bin_obj.id,
            "CDS": sum((contig_to_cds_count[c] for c in bin_obj.contigs if c in contig_to_cds_count)),
            "AALength": sum((contig_to_aa_length[c] for c in bin_obj.contigs if c in contig_to_aa_length)),
        }

        bin_aa_counter = Counter()
        for contig in bin_obj.contigs:
            if contig in contig_to_aa_counter:
                bin_aa_counter += contig_to_aa_counter[contig]

        bin_metadata.update(dict(bin_aa_counter))
        bin_metadata_list.append(bin_metadata)

    metadata_df = pd.DataFrame(bin_metadata_list, columns=["Name"] + metadata_order)

    metadata_df = metadata_df.fillna(0)

    metadata_df = metadata_df.astype({col: int for col in metadata_order})

    metadata_df = metadata_df.set_index("Name", drop=False)
    return metadata_df

def get_diamond_feature_per_bin_df(bins: List, contig_to_kegg_counter: Dict[str, Counter]) -> Tuple[pd.DataFrame, int]:
    """
    Generate a DataFrame containing Diamond feature counts per bin and completeness information for pathways, categories, and modules.

    :param bins: A list of bin objects.
    :param contig_to_kegg_counter: A dictionary mapping contig names to KEGG annotation counters.
    :type bins: List
    :type contig_to_kegg_counter: Dict[str, Counter]
    :return: A tuple containing the DataFrame and the number of default KEGG orthologs (KOs).
    :rtype: Tuple[pd.DataFrame, int]
    """
    KeggCalc = keggData.KeggCalculator()
    defaultKOs = KeggCalc.return_default_values_from_category("KO_Genes")

    bin_to_ko_counter = {}
    for bin_obj in bins:
        bin_ko_counter = Counter()
        for contig in bin_obj.contigs:
            try:
                bin_ko_counter += contig_to_kegg_counter[contig]
            except KeyError:
                # No KO annotation found in this contig
                continue

        bin_to_ko_counter[bin_obj.id] = bin_ko_counter

    ko_count_per_bin_df = pd.DataFrame(bin_to_ko_counter, index=defaultKOs).transpose().fillna(0)
    ko_count_per_bin_df = ko_count_per_bin_df.astype(int)
    ko_count_per_bin_df["Name"] = ko_count_per_bin_df.index

    logging.info("Calculating completeness of pathways and modules.")
    logging.debug("Calculating pathway completeness information")
    KO_pathways = KeggCalc.calculate_KO_group("KO_Pathways", ko_count_per_bin_df.copy())

    logging.debug("Calculating category completeness information")
    KO_categories = KeggCalc.calculate_KO_group("KO_Categories", ko_count_per_bin_df.copy())

    logging.debug("Calculating module completeness information")
    KO_modules = KeggCalc.calculate_module_completeness(ko_count_per_bin_df.copy())

    diamond_complete_results = pd.concat([ko_count_per_bin_df, KO_pathways, KO_modules, KO_categories], axis=1)

    return diamond_complete_results, len(defaultKOs)


def compute_N50(list_of_lengths) -> int:
    """
    Calculate N50 for a sequence of numbers.

    :param list_of_lengths: List of numbers.
    :param list_of_lengths: list
    :return: N50 value.
    """
    list_of_lengths = sorted(list_of_lengths)
    sum_len = sum(list_of_lengths)

    cum_length = 0
    length = 0
    for length in list_of_lengths:
        if cum_length + length >= sum_len / 2:
            return length
        cum_length += length
    return length

def add_bin_size_and_N50(bins: Iterable[Bin], contig_to_size: Dict[str,int]):
    """
    Add bin size and N50 to a list of bin objects.

    :param bins: List of bin objects.
    :param contig_to_size: Dictionary mapping contig names to their sizes.
    """
    for bin_obj in bins:
        lengths = [contig_to_size[c] for c in bin_obj.contigs]
        n50 = compute_N50(lengths)

        bin_obj.add_length(sum(lengths))
        bin_obj.add_N50(n50)


def add_bin_metrics(bins: List, contig_info: Dict, contamination_weight: float, threads: int = 1):
    """
    Add metrics to a list of bins.

    :param bins: List of bin objects.
    :param contig_info: Dictionary containing contig information.
    :param contamination_weight: Weight for contamination assessment.
    :param threads: Number of threads for parallel processing (default is 1).

    :return: List of processed bin objects.
    """
    postProcessor = modelPostprocessing.modelProcessor(threads)

    contig_to_kegg_counter = contig_info["contig_to_kegg_counter"]
    contig_to_cds_count = contig_info["contig_to_cds_count"]
    contig_to_aa_counter = contig_info["contig_to_aa_counter"]
    contig_to_aa_length = contig_info["contig_to_aa_length"]
    contig_to_length = contig_info["contig_to_length"]

    logging.info("Assess bin length and N50")

    add_bin_size_and_N50(bins, contig_to_length)

    logging.info("Assess bin quality")
    assess_bins_quality_by_chunk(
        bins,
        contig_to_kegg_counter,
        contig_to_cds_count,
        contig_to_aa_counter,
        contig_to_aa_length,
        contamination_weight,
        postProcessor,
    )
    return bins


def chunks(iterable: Iterable, size: int) -> Iterator[Tuple]:
    """
    Generate adjacent chunks of data from an iterable.

    :param iterable: The iterable to be divided into chunks.
    :param size: The size of each chunk.
    :return: An iterator that produces tuples of elements in chunks.
    """
    it = iter(iterable)
    return iter(lambda: tuple(islice(it, size)), ())


def assess_bins_quality_by_chunk(bins: List,
    contig_to_kegg_counter: Dict,
    contig_to_cds_count: Dict,
    contig_to_aa_counter: Dict,
    contig_to_aa_length: Dict,
    contamination_weight: float,
    postProcessor:modelPostprocessing.modelProcessor = None,
    threads: int = 1,
    chunk_size: int = 2500):
    """
    Assess the quality of bins in chunks.

    This function assesses the quality of bins in chunks to improve processing efficiency.

    :param bins: List of bin objects.
    :param contig_to_kegg_counter: Dictionary mapping contig names to KEGG counters.
    :param contig_to_cds_count: Dictionary mapping contig names to CDS counts.
    :param contig_to_aa_counter: Dictionary mapping contig names to amino acid counters.
    :param contig_to_aa_length: Dictionary mapping contig names to amino acid lengths.
    :param contamination_weight: Weight for contamination assessment.
    :param postProcessor: post-processor from checkm2
    :param threads: Number of threads for parallel processing (default is 1).
    :param chunk_size: The size of each chunk. 
    """

    for i, chunk_bins_iter in enumerate(chunks(bins, chunk_size)):
        chunk_bins = set(chunk_bins_iter)
        logging.debug(f"chunk {i}: assessing quality of {len(chunk_bins)}")
        assess_bins_quality(
            bins=chunk_bins,
            contig_to_kegg_counter= contig_to_kegg_counter,
            contig_to_cds_count=contig_to_cds_count,
            contig_to_aa_counter=contig_to_aa_counter,
            contig_to_aa_length=contig_to_aa_length,
            contamination_weight=contamination_weight,
            postProcessor=postProcessor,
            threads=threads
        )

def assess_bins_quality(
    bins: List,
    contig_to_kegg_counter: Dict,
    contig_to_cds_count: Dict,
    contig_to_aa_counter: Dict,
    contig_to_aa_length: Dict,
    contamination_weight: float,
    postProcessor: modelPostprocessing.modelProcessor = None,
    threads: int = 1,):
    """
    Assess the quality of bins.

    This function assesses the quality of bins based on various criteria and assigns completeness and contamination scores. 
    This code is taken from checkm2 and adjusted

    :param bins: List of bin objects.
    :param contig_to_kegg_counter: Dictionary mapping contig names to KEGG counters.
    :param contig_to_cds_count: Dictionary mapping contig names to CDS counts.
    :param contig_to_aa_counter: Dictionary mapping contig names to amino acid counters.
    :param contig_to_aa_length: Dictionary mapping contig names to amino acid lengths.
    :param contamination_weight: Weight for contamination assessment.
    :param postProcessor: A post-processor from checkm2
    :param threads: Number of threads for parallel processing (default is 1).
    """
    if postProcessor is None:
        postProcessor = modelPostprocessing.modelProcessor(threads)

    metadata_df = get_bins_metadata_df(bins, contig_to_cds_count, contig_to_aa_counter, contig_to_aa_length)

    diamond_complete_results, ko_list_length = get_diamond_feature_per_bin_df(bins, contig_to_kegg_counter)
    diamond_complete_results = diamond_complete_results.drop(columns=["Name"])

    feature_vectors = pd.concat([metadata_df, diamond_complete_results], axis=1)
    feature_vectors = feature_vectors.sort_values(by="Name")

    # 4: Call general model & specific models and derive predictions"""
    modelProc = modelProcessing.modelProcessor(threads)

    vector_array = feature_vectors.iloc[:, 1:].values.astype(np.float)

    logging.info("Predicting completeness and contamination using the general model.")
    general_results_comp, general_results_cont = modelProc.run_prediction_general(vector_array)

    logging.info("Predicting completeness using the specific model.")
    specific_model_vector_len = (
        ko_list_length + len(metadata_df.columns)
    ) - 1

    # also retrieve scaled data for CSM calculations
    specific_results_comp, scaled_features = modelProc.run_prediction_specific(vector_array, specific_model_vector_len)

    logging.info("Using cosine similarity to reference data to select an appropriate predictor model.")

    final_comp, final_cont, models_chosen, csm_array = postProcessor.calculate_general_specific_ratio(
        vector_array[:, 20], scaled_features, general_results_comp, general_results_cont, specific_results_comp
    )

    final_results = feature_vectors[["Name"]].copy()
    final_results["Completeness"] = np.round(final_comp, 2)
    final_results["Contamination"] = np.round(final_cont, 2)

    for bin_obj in bins:
        completeness = final_results.loc[bin_obj.id, "Completeness"]
        contamination = final_results.loc[bin_obj.id, "Contamination"]

        bin_obj.add_quality(completeness, contamination, contamination_weight)
